# Copyright (C) 2008, 2009, 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for branch implementations - test reconcile() functionality"""

from breezy import errors, reconcile
from breezy.bzr.branch import BzrBranch
from breezy.symbol_versioning import deprecated_in
from breezy.tests import TestNotApplicable
from breezy.tests.per_branch import TestCaseWithBranch


class TestBranchReconcile(TestCaseWithBranch):

    def test_reconcile_fixes_invalid_revhistory(self):
        if not isinstance(self.branch_format, BzrBranch):
            raise TestNotApplicable("test only applies to bzr formats")
        # Different formats have different ways of handling invalid revision
        # histories, so the setup portion is customized
        tree = self.make_branch_and_tree('test')
        r1 = tree.commit('one')
        r2 = tree.commit('two')
        r3 = tree.commit('three')
        r4 = tree.commit('four')
        # create an alternate branch
        tree.set_parent_ids([r1])
        tree.branch.set_last_revision_info(1, r1)
        r2b = tree.commit('two-b')

        # now go back and merge the commit
        tree.set_parent_ids([r4, r2b])
        tree.branch.set_last_revision_info(4, r4)

        r5 = tree.commit('five')
        # Now, try to set an invalid history
        try:
            self.applyDeprecated(deprecated_in((2, 4, 0)),
                                 tree.branch.set_revision_history, [r1, r2b, r5])
            if tree.branch.last_revision_info() != (3, r5):
                # RemoteBranch silently corrects an impossible revision
                # history given to set_revision_history.  It can be tricked
                # with set_last_revision_info though.
                tree.branch.set_last_revision_info(3, r5)
        except errors.NotLefthandHistory:
            # Branch5 allows set_revision_history to be wrong
            # Branch6 raises NotLefthandHistory, but we can force bogus stuff
            # with set_last_revision_info
            tree.branch.set_last_revision_info(3, r5)

        self.assertEqual((3, r5), tree.branch.last_revision_info())
        reconciler = tree.branch.reconcile()
        self.assertEqual((5, r5), tree.branch.last_revision_info())
        self.assertIs(True, reconciler.fixed_history)

    def test_reconcile_returns_reconciler(self):
        a_branch = self.make_branch('a_branch')
        result = a_branch.reconcile()
        self.assertIsInstance(result, reconcile.ReconcileResult)
        # No history to fix
        self.assertIs(False, getattr(result, 'fixed_history', False))

    def test_reconcile_supports_thorough(self):
        a_branch = self.make_branch('a_branch')
        a_branch.reconcile(thorough=False)
        a_branch.reconcile(thorough=True)

    def test_reconcile_handles_ghosts_in_revhistory(self):
        tree = self.make_branch_and_tree('test')
        if not tree.branch.repository._format.supports_ghosts:
            raise TestNotApplicable(
                "repository format does not support ghosts")
        tree.set_parent_ids([b"spooky"], allow_leftmost_as_ghost=True)
        r1 = tree.commit('one')
        r2 = tree.commit('two')
        tree.branch.set_last_revision_info(2, r2)

        reconciler = tree.branch.reconcile()
        self.assertEqual(r2, tree.branch.last_revision())



# Copyright (C) 2020 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from datetime import datetime

from breezy.plugins.gitlab.forge import (NotGitLabUrl, NotMergeRequestUrl,
                                         parse_gitlab_merge_request_url,
                                         parse_timestring)
from breezy.tests import TestCase


class ParseGitLabMergeRequestUrlTests(TestCase):

    def test_invalid(self):
        self.assertRaises(
            NotMergeRequestUrl, parse_gitlab_merge_request_url,
            "https://salsa.debian.org/")
        self.assertRaises(
            NotGitLabUrl, parse_gitlab_merge_request_url,
            "bzr://salsa.debian.org/")
        self.assertRaises(
            NotGitLabUrl, parse_gitlab_merge_request_url,
            "https:///salsa.debian.org/")
        self.assertRaises(
            NotMergeRequestUrl, parse_gitlab_merge_request_url,
            "https://salsa.debian.org/jelmer/salsa")

    def test_old_style(self):
        self.assertEqual(
            ('salsa.debian.org', 'jelmer/salsa', 4),
            parse_gitlab_merge_request_url(
                'https://salsa.debian.org/jelmer/salsa/merge_requests/4'))

    def test_new_style(self):
        self.assertEqual(
            ('salsa.debian.org', 'jelmer/salsa', 4),
            parse_gitlab_merge_request_url(
                'https://salsa.debian.org/jelmer/salsa/-/merge_requests/4'))


class ParseTimestringTests(TestCase):

    def test_simple(self):
        self.assertEqual(
            datetime(2018, 9, 7, 11, 16, 17, 520000),
            parse_timestring("2018-09-07T11:16:17.520Z"))



# Copyright (C) 2006-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test locks across all branch implemenations"""

from breezy import errors, tests
from breezy.tests import lock_helpers, per_branch
from breezy.tests.matchers import *


class TestBranchLocking(per_branch.TestCaseWithBranch):

    def setUp(self):
        super().setUp()
        self.reduceLockdirTimeout()

    def get_instrumented_branch(self):
        """Get a Branch object which has been instrumented"""
        # TODO: jam 20060630 It may be that not all formats have a
        # 'control_files' member. So we should fail gracefully if
        # not there. But assuming it has them lets us test the exact
        # lock/unlock order.
        self.locks = []
        b = lock_helpers.LockWrapper(self.locks, self.get_branch(), 'b')
        b.repository = lock_helpers.LockWrapper(self.locks, b.repository, 'r')
        bcf = getattr(b, "control_files", None)
        rcf = getattr(b.repository, 'control_files', None)
        if rcf is None:
            self.combined_branch = False
        else:
            # Look out for branch types that reuse their control files
            self.combined_control = bcf is rcf and bcf is not None
        try:
            b.control_files = lock_helpers.LockWrapper(
                self.locks, b.control_files, 'bc')
        except AttributeError:
            # RemoteBranch seems to trigger this.
            raise tests.TestSkipped(
                'Could not instrument branch control files.')
        if self.combined_control:
            # instrument the repository control files too to ensure its worked
            # with correctly. When they are not shared, we trust the repository
            # API and only instrument the repository itself.
            b.repository.control_files = lock_helpers.LockWrapper(
                self.locks, b.repository.control_files, 'rc')
        return b

    def test_01_lock_read(self):
        # Test that locking occurs in the correct order
        b = self.get_instrumented_branch()

        self.assertFalse(b.is_locked())
        self.assertFalse(b.repository.is_locked())
        b.lock_read()
        try:
            self.assertTrue(b.is_locked())
            self.assertTrue(b.repository.is_locked())
        finally:
            b.unlock()
        self.assertFalse(b.is_locked())
        self.assertFalse(b.repository.is_locked())

        if self.combined_control:
            self.assertEqual([('b', 'lr', True),
                              ('r', 'lr', True),
                              ('rc', 'lr', True),
                              ('bc', 'lr', True),
                              ('b', 'ul', True),
                              ('bc', 'ul', True),
                              ('r', 'ul', True),
                              ('rc', 'ul', True),
                              ], self.locks)
        else:
            self.assertEqual([('b', 'lr', True),
                              ('r', 'lr', True),
                              ('bc', 'lr', True),
                              ('b', 'ul', True),
                              ('bc', 'ul', True),
                              ('r', 'ul', True),
                              ], self.locks)

    def test_02_lock_write(self):
        # Test that locking occurs in the correct order
        b = self.get_instrumented_branch()

        self.assertFalse(b.is_locked())
        self.assertFalse(b.repository.is_locked())
        b.lock_write()
        try:
            self.assertTrue(b.is_locked())
            self.assertTrue(b.repository.is_locked())
        finally:
            b.unlock()
        self.assertFalse(b.is_locked())
        self.assertFalse(b.repository.is_locked())

        if self.combined_control:
            self.assertEqual([('b', 'lw', True),
                              ('r', 'lw', True),
                              ('rc', 'lw', True),
                              ('bc', 'lw', True),
                              ('b', 'ul', True),
                              ('bc', 'ul', True),
                              ('r', 'ul', True),
                              ('rc', 'ul', True),
                              ], self.locks)
        else:
            self.assertEqual([('b', 'lw', True),
                              ('r', 'lw', True),
                              ('bc', 'lw', True),
                              ('b', 'ul', True),
                              ('bc', 'ul', True),
                              ('r', 'ul', True),
                              ], self.locks)

    def test_03_lock_fail_unlock_repo(self):
        # Make sure branch.unlock() is called, even if there is a
        # failure while unlocking the repository.
        b = self.get_instrumented_branch()
        b.repository.disable_unlock()

        self.assertFalse(b.is_locked())
        self.assertFalse(b.repository.is_locked())
        b.lock_write()
        try:
            self.assertTrue(b.is_locked())
            self.assertTrue(b.repository.is_locked())
            self.assertLogsError(lock_helpers.TestPreventLocking, b.unlock)
            if self.combined_control:
                self.assertTrue(b.is_locked())
            else:
                self.assertFalse(b.is_locked())
            self.assertTrue(b.repository.is_locked())

            # We unlock the branch control files, even if
            # we fail to unlock the repository
            if self.combined_control:
                self.assertEqual([('b', 'lw', True),
                                  ('r', 'lw', True),
                                  ('rc', 'lw', True),
                                  ('bc', 'lw', True),
                                  ('b', 'ul', True),
                                  ('bc', 'ul', True),
                                  ('r', 'ul', False),
                                  ], self.locks)
            else:
                self.assertEqual([('b', 'lw', True),
                                  ('r', 'lw', True),
                                  ('bc', 'lw', True),
                                  ('b', 'ul', True),
                                  ('bc', 'ul', True),
                                  ('r', 'ul', False),
                                  ], self.locks)

        finally:
            # For cleanup purposes, make sure we are unlocked
            b.repository._other.unlock()

    def test_04_lock_fail_unlock_control(self):
        # Make sure repository.unlock() is not called, if we fail to unlock
        # self leaving ourselves still locked, so that attempts to recover
        # don't encounter an unlocked repository.
        b = self.get_instrumented_branch()
        b.control_files.disable_unlock()

        self.assertFalse(b.is_locked())
        self.assertFalse(b.repository.is_locked())
        b.lock_write()
        try:
            self.assertTrue(b.is_locked())
            self.assertTrue(b.repository.is_locked())
            self.assertLogsError(lock_helpers.TestPreventLocking, b.unlock)
            self.assertTrue(b.is_locked())
            self.assertTrue(b.repository.is_locked())

            # We unlock the repository even if
            # we fail to unlock the control files
            if self.combined_control:
                self.assertEqual([('b', 'lw', True),
                                  ('r', 'lw', True),
                                  ('rc', 'lw', True),
                                  ('bc', 'lw', True),
                                  ('b', 'ul', True),
                                  ('bc', 'ul', False),
                                  ('r', 'ul', True),
                                  ('rc', 'ul', True),
                                  ], self.locks)
            else:
                self.assertEqual([('b', 'lw', True),
                                  ('r', 'lw', True),
                                  ('bc', 'lw', True),
                                  ('b', 'ul', True),
                                  ('bc', 'ul', False),
                                  ], self.locks)

        finally:
            # For cleanup purposes, make sure we are unlocked
            b.control_files._other.unlock()

    def test_05_lock_read_fail_repo(self):
        # Test that the branch is not locked if it cannot lock the repository
        b = self.get_instrumented_branch()
        b.repository.disable_lock_read()

        self.assertRaises(lock_helpers.TestPreventLocking, b.lock_read)
        self.assertFalse(b.is_locked())
        self.assertFalse(b.repository.is_locked())

        self.assertEqual([('b', 'lr', True),
                          ('r', 'lr', False),
                          ], self.locks)

    def test_06_lock_write_fail_repo(self):
        # Test that the branch is not locked if it cannot lock the repository
        b = self.get_instrumented_branch()
        b.repository.disable_lock_write()

        self.assertRaises(lock_helpers.TestPreventLocking, b.lock_write)
        self.assertFalse(b.is_locked())
        self.assertFalse(b.repository.is_locked())

        self.assertEqual([('b', 'lw', True),
                          ('r', 'lw', False),
                          ], self.locks)

    def test_07_lock_read_fail_control(self):
        # Test the repository is unlocked if we can't lock self
        b = self.get_instrumented_branch()
        b.control_files.disable_lock_read()

        self.assertRaises(lock_helpers.TestPreventLocking, b.lock_read)
        self.assertFalse(b.is_locked())
        self.assertFalse(b.repository.is_locked())

        if self.combined_control:
            self.assertEqual([('b', 'lr', True),
                              ('r', 'lr', True),
                              ('rc', 'lr', True),
                              ('bc', 'lr', False),
                              ('r', 'ul', True),
                              ('rc', 'ul', True),
                              ], self.locks)
        else:
            self.assertEqual([('b', 'lr', True),
                              ('r', 'lr', True),
                              ('bc', 'lr', False),
                              ('r', 'ul', True),
                              ], self.locks)

    def test_08_lock_write_fail_control(self):
        # Test the repository is unlocked if we can't lock self
        b = self.get_instrumented_branch()
        b.control_files.disable_lock_write()

        self.assertRaises(lock_helpers.TestPreventLocking, b.lock_write)
        self.assertFalse(b.is_locked())
        self.assertFalse(b.repository.is_locked())
        if self.combined_control:
            self.assertEqual([('b', 'lw', True),
                              ('r', 'lw', True),
                              ('rc', 'lw', True),
                              ('bc', 'lw', False),
                              ('r', 'ul', True),
                              ('rc', 'ul', True),
                              ], self.locks)
        else:
            self.assertEqual([('b', 'lw', True),
                              ('r', 'lw', True),
                              ('bc', 'lw', False),
                              ('r', 'ul', True),
                              ], self.locks)

    def test_lock_write_returns_None_refuses_token(self):
        branch = self.make_branch('b')
        with branch.lock_write() as lock:
            if lock.token is not None:
                # This test does not apply, because this lockable supports
                # tokens.
                return
            self.assertRaises(errors.TokenLockingNotSupported,
                              branch.lock_write, token='token')

    def test_reentering_lock_write_raises_on_token_mismatch(self):
        branch = self.make_branch('b')
        with branch.lock_write() as lock:
            if lock.token is None:
                # This test does not apply, because this lockable refuses
                # tokens.
                return
            different_branch_token = lock.token + b'xxx'
            # Re-using the same lockable instance with a different branch token
            # will raise TokenMismatch.
            self.assertRaises(errors.TokenMismatch,
                              branch.lock_write,
                              token=different_branch_token)

    def test_lock_write_with_nonmatching_token(self):
        branch = self.make_branch('b')
        with branch.lock_write() as lock:
            if lock.token is None:
                # This test does not apply, because this branch refuses
                # tokens.
                return
            different_branch_token = lock.token + b'xxx'

            new_branch = branch.controldir.open_branch()
            # We only want to test the relocking abilities of branch, so use the
            # existing repository object which is already locked.
            new_branch.repository = branch.repository
            self.assertRaises(errors.TokenMismatch,
                              new_branch.lock_write,
                              token=different_branch_token)

    def test_lock_write_with_matching_token(self):
        """Test that a branch can be locked with a token, if it is already
        locked by that token."""
        branch = self.make_branch('b')
        with branch.lock_write() as lock:
            if lock.token is None:
                # This test does not apply, because this branch refuses tokens.
                return
            # The same instance will accept a second lock_write if the specified
            # token matches.
            branch.lock_write(token=lock.token)
            branch.unlock()
            # Calling lock_write on a new instance for the same lockable will
            # also succeed.
            new_branch = branch.controldir.open_branch()
            # We only want to test the relocking abilities of branch, so use the
            # existing repository object which is already locked.
            new_branch.repository = branch.repository
            new_branch.lock_write(token=lock.token)
            new_branch.unlock()

    def test_unlock_after_lock_write_with_token(self):
        # If lock_write did not physically acquire the lock (because it was
        # passed some tokens), then unlock should not physically release it.
        branch = self.make_branch('b')
        with branch.lock_write() as lock:
            if lock.token is None:
                # This test does not apply, because this lockable refuses
                # tokens.
                return
            new_branch = branch.controldir.open_branch()
            # We only want to test the relocking abilities of branch, so use the
            # existing repository object which is already locked.
            new_branch.repository = branch.repository
            new_branch.lock_write(token=lock.token)
            new_branch.unlock()
            self.assertTrue(branch.get_physical_lock_status())  # XXX

    def test_lock_write_with_token_fails_when_unlocked(self):
        # First, lock and then unlock to get superficially valid tokens.  This
        # mimics a likely programming error, where a caller accidentally tries
        # to lock with a token that is no longer valid (because the original
        # lock was released).
        branch = self.make_branch('b')
        token = branch.lock_write().token
        branch.unlock()
        if token is None:
            # This test does not apply, because this lockable refuses
            # tokens.
            return
        self.assertRaises(errors.TokenMismatch, branch.lock_write, token=token)

    def test_lock_write_reenter_with_token(self):
        branch = self.make_branch('b')
        with branch.lock_write() as lock:
            if lock.token is None:
                # This test does not apply, because this lockable refuses
                # tokens.
                return
            # Relock with a token.
            branch.lock_write(token=lock.token)
            branch.unlock()
        # The lock should be unlocked on disk.  Verify that with a new lock
        # instance.
        new_branch = branch.controldir.open_branch()
        # Calling lock_write now should work, rather than raise LockContention.
        new_branch.lock_write()
        new_branch.unlock()

    def test_leave_lock_in_place(self):
        branch = self.make_branch('b')
        # Lock the branch, then use leave_lock_in_place so that when we
        # unlock the branch the lock is still held on disk.
        token = None
        with branch.lock_write() as lock:
            token = lock.token
            if lock.token is None:
                # This test does not apply, because this repository refuses
                # lock tokens.
                self.assertRaises(NotImplementedError,
                                  branch.leave_lock_in_place)
                return
            branch.leave_lock_in_place()
        # We should be unable to relock the repo.
        self.assertRaises(errors.LockContention, branch.lock_write)
        # Cleanup
        branch.lock_write(token)
        branch.dont_leave_lock_in_place()
        branch.unlock()

    def test_dont_leave_lock_in_place(self):
        branch = self.make_branch('b')
        # Create a lock on disk.
        token = branch.lock_write().token
        try:
            if token is None:
                # This test does not apply, because this branch refuses lock
                # tokens.
                self.assertRaises(NotImplementedError,
                                  branch.dont_leave_lock_in_place)
                return
            try:
                branch.leave_lock_in_place()
            except NotImplementedError:
                # This branch doesn't support this API.
                return
            try:
                branch.repository.leave_lock_in_place()
            except NotImplementedError:
                # This repo doesn't support leaving locks around,
                # assume it is essentially lock-free.
                repo_token = None
            else:
                repo_token = branch.repository.lock_write()
                branch.repository.unlock()
        finally:
            branch.unlock()
        # Reacquire the lock (with a different branch object) by using the
        # tokens.
        new_branch = branch.controldir.open_branch()
        if repo_token is not None:
            # We have to explicitly lock the repository first.
            new_branch.repository.lock_write(token=repo_token)
        new_branch.lock_write(token=token)
        if repo_token is not None:
            # Now we don't need our own repository lock anymore (the branch is
            # holding it for us).
            new_branch.repository.unlock()
        # Call dont_leave_lock_in_place, so that the lock will be released by
        # this instance, even though the lock wasn't originally acquired by it.
        new_branch.dont_leave_lock_in_place()
        if repo_token is not None:
            new_branch.repository.dont_leave_lock_in_place()
        new_branch.unlock()
        # Now the branch (and repository) is unlocked.  Test this by locking it
        # without tokens.
        branch.lock_write()
        branch.unlock()

    def test_lock_read_then_unlock(self):
        # Calling lock_read then unlocking should work without errors.
        branch = self.make_branch('b')
        branch.lock_read()
        branch.unlock()

    def test_lock_read_context_manager(self):
        # Calling lock_read then unlocking should work without errors.
        branch = self.make_branch('b')
        self.assertFalse(branch.is_locked())
        with branch.lock_read():
            self.assertTrue(branch.is_locked())

    def test_lock_read_returns_unlockable(self):
        branch = self.make_branch('b')
        self.assertThat(branch.lock_read, ReturnsUnlockable(branch))

    def test_lock_write_locks_repo_too(self):
        branch = self.make_branch('b')
        branch = branch.controldir.open_branch()
        branch.lock_write()
        try:
            # The branch should have asked the repository to lock.
            self.assertTrue(branch.repository.is_write_locked())
            # Does the repository type actually lock?
            if not branch.repository.get_physical_lock_status():
                # The test was successfully applied, so it was applicable.
                return
            # Now the branch.repository is physically locked, so we can't lock
            # it with a new repository instance.
            new_repo = branch.controldir.open_repository()
            self.assertRaises(errors.LockContention, new_repo.lock_write)
            # We can call lock_write on the original repository object though,
            # because it is already locked.
            branch.repository.lock_write()
            branch.repository.unlock()
        finally:
            branch.unlock()

    def test_lock_write_returns_unlockable(self):
        branch = self.make_branch('b')
        self.assertThat(branch.lock_write, ReturnsUnlockable(branch))

    def test_lock_write_raises_in_lock_read(self):
        branch = self.make_branch('b')
        branch.lock_read()
        self.addCleanup(branch.unlock)
        err = self.assertRaises(errors.ReadOnlyError, branch.lock_write)

    def test_lock_and_unlock_leaves_repo_unlocked(self):
        branch = self.make_branch('b')
        branch.lock_write()
        branch.unlock()
        self.assertRaises(errors.LockNotHeld, branch.repository.unlock)



# Copyright (C) 2020 Jelmer Vernooij
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from ..memorybranch import MemoryBranch
from . import TestCaseWithTransport


class MemoryBranchTests(TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        self.tree = self.make_branch_and_tree('.')
        self.revid1 = self.tree.commit('rev1')
        self.revid2 = self.tree.commit('rev2')
        self.branch = MemoryBranch(
            self.tree.branch.repository, (2, self.revid2))

    def test_last_revision_info(self):
        self.assertEqual((2, self.revid2), self.branch.last_revision_info())

    def test_last_revision(self):
        self.assertEqual(self.revid2, self.branch.last_revision())

    def test_revno(self):
        self.assertEqual(2, self.branch.revno())

    def test_get_rev_id(self):
        self.assertEqual(self.revid1, self.branch.get_rev_id(1))

    def test_revision_id_to_revno(self):
        self.assertEqual(2, self.branch.revision_id_to_revno(self.revid2))
        self.assertEqual(1, self.branch.revision_id_to_revno(self.revid1))



# Copyright (C) 2007-2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Commands behaviour tests for bzr.

Test the internal behaviour of the commands (the blackbox tests are intended to
test the usage of the commands).
"""

# FIXME: If the separation described above from the blackbox tests is not worth
# it, all the tests defined below should be moved to blackbox instead.


def load_tests(loader, basic_tests, pattern):
    suite = loader.suiteClass()
    # add the tests for this module
    suite.addTests(basic_tests)

    testmod_names = [
        'breezy.tests.commands.test_branch',
        'breezy.tests.commands.test_cat',
        'breezy.tests.commands.test_checkout',
        'breezy.tests.commands.test_commit',
        'breezy.tests.commands.test_init',
        'breezy.tests.commands.test_init_repository',
        'breezy.tests.commands.test_merge',
        'breezy.tests.commands.test_missing',
        'breezy.tests.commands.test_pull',
        'breezy.tests.commands.test_push',
        'breezy.tests.commands.test_update',
        'breezy.tests.commands.test_revert',
        ]
    # add the tests for the sub modules
    suite.addTests(loader.loadTestsFromModuleNames(testmod_names))

    return suite



# Copyright (C) 2005 Canonical Ltd
#   Authors: Robert Collins <robert.collins@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""This module provides an IdentityMap."""

from . import errors


class IdentityMap:
    """An in memory map from object id to instance.

    An IdentityMap maps from keys to single instances of objects in memory.
    We have explicit calls on the map for the root of each inheritance tree
    that is store in the map. Look for find_CLASS and add_CLASS methods.
    """

    def add_weave(self, id, weave):
        """Add weave to the map with a given id."""
        if self._weave_key(id) in self._map:
            raise errors.BzrError('weave %s already in the identity map' % id)
        self._map[self._weave_key(id)] = weave
        self._reverse_map[weave] = self._weave_key(id)

    def find_weave(self, id):
        """Return the weave for 'id', or None if it is not present."""
        return self._map.get(self._weave_key(id), None)

    def __init__(self):
        super().__init__()
        self._map = {}
        self._reverse_map = {}

    def remove_object(self, an_object):
        """Remove object from map."""
        if isinstance(an_object, list):
            raise KeyError('%r not in identity map' % an_object)
        else:
            self._map.pop(self._reverse_map[an_object])
            self._reverse_map.pop(an_object)

    def _weave_key(self, id):
        """Return the key for a weaves id."""
        return "weave-" + id


class NullIdentityMap:
    """A pretend in memory map from object id to instance.

    A NullIdentityMap is an Identity map that does not store anything in it.
    """

    def add_weave(self, id, weave):
        """See IdentityMap.add_weave."""

    def find_weave(self, id):
        """See IdentityMap.find_weave."""
        return None



# Copyright (C) 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Black-box tests for brz lookup-revision.
"""

from breezy import tests


class TestLookupRevision(tests.TestCaseWithTransport):

    def test_lookup_revison_directory(self):
        """Test --directory option"""
        tree = self.make_branch_and_tree('a')
        tree.commit('This revision', rev_id=b'abcd')
        out, err = self.run_bzr(['lookup-revision', '-d', 'a', '1'])
        self.assertEqual('abcd\n', out)
        self.assertEqual('', err)



# this is copied from the lsprof distro because somehow
# it is not installed by distutils
# I made one modification to profile so that it returns a pair
# instead of just the Stats object

import _thread
import codecs
import operator
import os
import pickle
import sys
import threading
from typing import Dict, TextIO

from _lsprof import Profiler, profiler_entry

from . import errors

__all__ = ['profile', 'Stats']


def profile(f, *args, **kwds):
    """Run a function profile.

    Exceptions are not caught: If you need stats even when exceptions are to be
    raised, pass in a closure that will catch the exceptions and transform them
    appropriately for your driver function.

    Important caveat: only one profile can execute at a time. See BzrProfiler
    for details.

    :return: The functions return value and a stats object.
    """
    profiler = BzrProfiler()
    profiler.start()
    try:
        ret = f(*args, **kwds)
    finally:
        stats = profiler.stop()
    return ret, stats


class BzrProfiler:
    """Bzr utility wrapper around Profiler.

    For most uses the module level 'profile()' function will be suitable.
    However profiling when a simple wrapped function isn't available may
    be easier to accomplish using this class.

    To use it, create a BzrProfiler and call start() on it. Some arbitrary
    time later call stop() to stop profiling and retrieve the statistics
    from the code executed in the interim.

    Note that profiling involves a threading.Lock around the actual profiling.
    This is needed because profiling involves global manipulation of the python
    interpreter state. As such you cannot perform multiple profiles at once.
    Trying to do so will lock out the second profiler unless the global
    breezy.lsprof.BzrProfiler.profiler_block is set to 0. Setting it to 0 will
    cause profiling to fail rather than blocking.
    """

    profiler_block = 1
    """Serialise rather than failing to profile concurrent profile requests."""

    profiler_lock = threading.Lock()
    """Global lock used to serialise profiles."""

    def start(self):
        """Start profiling.

        This hooks into threading and will record all calls made until
        stop() is called.
        """
        self._g_threadmap = {}
        self.p = Profiler()
        permitted = self.__class__.profiler_lock.acquire(
            self.__class__.profiler_block)
        if not permitted:
            raise errors.InternalBzrError(msg="Already profiling something")
        try:
            self.p.enable(subcalls=True)
            threading.setprofile(self._thread_profile)
        except BaseException:
            self.__class__.profiler_lock.release()
            raise

    def stop(self):
        """Stop profiling.

        This unhooks from threading and cleans up the profiler, returning
        the gathered Stats object.

        :return: A breezy.lsprof.Stats object.
        """
        try:
            self.p.disable()
            for pp in self._g_threadmap.values():
                pp.disable()
            threading.setprofile(None)
            p = self.p
            self.p = None
            threads = {}
            for tid, pp in self._g_threadmap.items():
                threads[tid] = Stats(pp.getstats(), {})
            self._g_threadmap = None
            return Stats(p.getstats(), threads)
        finally:
            self.__class__.profiler_lock.release()

    def _thread_profile(self, f, *args, **kwds):
        # we lose the first profile point for a new thread in order to
        # trampoline a new Profile object into place
        thr = _thread.get_ident()
        self._g_threadmap[thr] = p = Profiler()
        # this overrides our sys.setprofile hook:
        p.enable(subcalls=True, builtins=True)


class Stats:
    """Wrapper around the collected data.

    A Stats instance is created when the profiler finishes. Normal
    usage is to use save() to write out the data to a file, or pprint()
    to write human-readable information to the command line.
    """

    def __init__(self, data, threads):
        self.data = data
        self.threads = threads

    def sort(self, crit="inlinetime", reverse=True):
        """Sort the data by the supplied critera.

        :param crit: the data attribute used as the sort key."""
        if crit not in profiler_entry.__dict__ or crit == 'code':
            raise ValueError("Can't sort by %s" % crit)

        key_func = operator.attrgetter(crit)
        self.data.sort(key=key_func, reverse=reverse)

        for e in self.data:
            if e.calls:
                e.calls.sort(key=key_func, reverse=reverse)

    def pprint(self, top=None, file=None):
        """Pretty-print the data as plain text for human consumption.

        :param top: only output the top n entries.
            The default value of None means output all data.
        :param file: the output file; if None, output will
            default to stdout."""
        if file is None:
            file = sys.stdout
        d = self.data
        if top is not None:
            d = d[:top]
        cols = "% 12s %12s %11.4f %11.4f   %s\n"
        hcols = "% 12s %12s %12s %12s %s\n"
        file.write(hcols % ("CallCount", "Recursive", "Total(ms)",
                            "Inline(ms)", "module:lineno(function)"))
        for e in d:
            file.write(cols % (e.callcount, e.reccallcount, e.totaltime,
                               e.inlinetime, label(e.code)))
            if e.calls:
                for se in e.calls:
                    file.write(cols % ("+%s" % se.callcount, se.reccallcount,
                                       se.totaltime, se.inlinetime,
                                       "+%s" % label(se.code)))

    def freeze(self):
        """Replace all references to code objects with string
        descriptions; this makes it possible to pickle the instance."""

        # this code is probably rather ickier than it needs to be!
        for i in range(len(self.data)):
            e = self.data[i]
            if not isinstance(e.code, str):
                self.data[i] = type(e)((label(e.code),) + e[1:])
            if e.calls:
                for j in range(len(e.calls)):
                    se = e.calls[j]
                    if not isinstance(se.code, str):
                        e.calls[j] = type(se)((label(se.code),) + se[1:])
        for s in self.threads.values():
            s.freeze()

    def calltree(self, file):
        """Output profiling data in calltree format (for KCacheGrind)."""
        _CallTreeFilter(self.data).output(file)

    def save(self, filename, format=None):
        """Save profiling data to a file.

        :param filename: the name of the output file
        :param format: 'txt' for a text representation;
            'callgrind' for calltree format;
            otherwise a pickled Python object. A format of None indicates
            that the format to use is to be found from the filename. If
            the name starts with callgrind.out, callgrind format is used
            otherwise the format is given by the filename extension.
        """
        if format is None:
            basename = os.path.basename(filename)
            if basename.startswith('callgrind.out'):
                format = "callgrind"
            else:
                ext = os.path.splitext(filename)[1]
                if len(ext) > 1:
                    format = ext[1:]
        with open(filename, 'wb') as outfile:
            if format == "callgrind":
                # The callgrind format states it is 'ASCII based':
                # <http://valgrind.org/docs/manual/cl-format.html>
                # But includes filenames so lets ignore and use UTF-8.
                self.calltree(codecs.getwriter('utf-8')(outfile))
            elif format == "txt":
                self.pprint(file=codecs.getwriter('utf-8')(outfile))
            else:
                self.freeze()
                pickle.dump(self, outfile, 2)


class _CallTreeFilter:
    """Converter of a Stats object to input suitable for KCacheGrind.

    This code is taken from http://ddaa.net/blog/python/lsprof-calltree
    with the changes made by J.P. Calderone and Itamar applied. Note that
    isinstance(code, str) needs to be used at times to determine if the code
    object is actually an external code object (with a filename, etc.) or
    a Python built-in.
    """

    out_file: TextIO

    def __init__(self, data):
        self.data = data

    def output(self, out_file):
        self.out_file = out_file
        out_file.write('events: Ticks\n')
        self._print_summary()
        for entry in self.data:
            self._entry(entry)

    def _print_summary(self):
        max_cost = 0
        for entry in self.data:
            totaltime = int(entry.totaltime * 1000)
            max_cost = max(max_cost, totaltime)
        self.out_file.write('summary: %d\n' % (max_cost,))

    def _entry(self, entry):
        out_file = self.out_file
        code = entry.code
        inlinetime = int(entry.inlinetime * 1000)
        if isinstance(code, str):
            out_file.write('fi=~\n')
        else:
            out_file.write('fi={}\n'.format(code.co_filename))
        out_file.write('fn={}\n'.format(label(code, True)))
        if isinstance(code, str):
            out_file.write('0  {}\n'.format(inlinetime))
        else:
            out_file.write('%d %d\n' % (code.co_firstlineno, inlinetime))
        # recursive calls are counted in entry.calls
        if entry.calls:
            calls = entry.calls
        else:
            calls = []
        if isinstance(code, str):
            lineno = 0
        else:
            lineno = code.co_firstlineno
        for subentry in calls:
            self._subentry(lineno, subentry)
        out_file.write('\n')

    def _subentry(self, lineno, subentry):
        out_file = self.out_file
        code = subentry.code
        totaltime = int(subentry.totaltime * 1000)
        if isinstance(code, str):
            out_file.write('cfi=~\n')
            out_file.write('cfn={}\n'.format(label(code, True)))
            out_file.write('calls=%d 0\n' % (subentry.callcount,))
        else:
            out_file.write('cfi={}\n'.format(code.co_filename))
            out_file.write('cfn={}\n'.format(label(code, True)))
            out_file.write('calls=%d %d\n' % (
                subentry.callcount, code.co_firstlineno))
        out_file.write('%d %d\n' % (lineno, totaltime))


_fn2mod: Dict[str, object] = {}


def label(code, calltree=False):
    if isinstance(code, str):
        return code
    try:
        mname = _fn2mod[code.co_filename]
    except KeyError:
        for k, v in sys.modules.items():
            if v is None:
                continue
            if getattr(v, '__file__', None) is None:
                continue
            if not isinstance(v.__file__, str):
                continue
            if v.__file__.startswith(code.co_filename):
                mname = _fn2mod[code.co_filename] = k
                break
        else:
            mname = _fn2mod[code.co_filename] = '<%s>' % code.co_filename
    if calltree:
        return '%s %s:%d' % (code.co_name, mname, code.co_firstlineno)
    else:
        return '%s:%d(%s)' % (mname, code.co_firstlineno, code.co_name)


def main():
    sys.argv = sys.argv[1:]
    if not sys.argv:
        sys.stderr.write("usage: lsprof.py <script> <arguments...>\n")
        sys.exit(2)
    import runpy
    result, stats = profile(runpy.run_path, sys.argv[0], run_name='__main__')
    stats.sort()
    stats.pprint()


if __name__ == '__main__':
    main()



# Copyright (C) 2010-2018 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Roundtripping support.

Bazaar stores more data than Git, which means that in order to preserve
a commit when it is pushed from Bazaar into Git we have to stash
that extra metadata somewhere.

There are two kinds of metadata relevant here:
 * per-file metadata (stored by revision+path)
  - usually stored per tree
 * per-revision metadata (stored by git commit id)

Bazaar revisions have the following information that is not
present in Git commits:
 * revision ids
 * revision properties
 * ghost parents

Tree content:
 * empty directories
 * path file ids
 * path last changed revisions [1]

 [1] path last changed revision information can usually
     be induced from the existing history, unless
     ghost revisions are involved.

This extra metadata is stored in so-called "supplements":
  * CommitSupplement
  * TreeSupplement
"""

from io import BytesIO

from .. import osutils


class CommitSupplement:
    """Supplement for a Bazaar revision roundtripped into Git.

    :ivar revision_id: Revision id, as string
    :ivar properties: Revision properties, as dictionary
    :ivar explicit_parent_ids: Parent ids (needed if there are ghosts)
    :ivar verifiers: Verifier information
    """

    revision_id = None

    explicit_parent_ids = None

    def __init__(self):
        self.properties = {}
        self.verifiers = {}

    def __nonzero__(self):
        return bool(self.revision_id or self.properties or
                    self.explicit_parent_ids)


class TreeSupplement:
    """Supplement for a Bazaar tree roundtripped into Git.

    This provides file ids (if they are different from the mapping default)
    and can provide text revisions.
    """


def parse_roundtripping_metadata(text):
    """Parse Bazaar roundtripping metadata."""
    ret = CommitSupplement()
    f = BytesIO(text)
    for l in f.readlines():
        (key, value) = l.split(b":", 1)
        if key == b"revision-id":
            ret.revision_id = value.strip()
        elif key == b"parent-ids":
            ret.explicit_parent_ids = tuple(value.strip().split(b" "))
        elif key == b"testament3-sha1":
            ret.verifiers[b"testament3-sha1"] = value.strip()
        elif key.startswith(b"property-"):
            name = key[len(b"property-"):]
            if name not in ret.properties:
                ret.properties[name] = value[1:].rstrip(b"\n")
            else:
                ret.properties[name] += b"\n" + value[1:].rstrip(b"\n")
        else:
            raise ValueError
    return ret


def generate_roundtripping_metadata(metadata, encoding):
    """Serialize the roundtripping metadata.

    :param metadata: A `CommitSupplement` instance
    :return: String with revision metadata
    """
    lines = []
    if metadata.revision_id:
        lines.append(b"revision-id: %s\n" % metadata.revision_id)
    if metadata.explicit_parent_ids:
        lines.append(b"parent-ids: %s\n" %
                     b" ".join(metadata.explicit_parent_ids))
    for key in sorted(metadata.properties.keys()):
        for l in metadata.properties[key].split(b"\n"):
            lines.append(b"property-%s: %s\n" % (key, osutils.safe_utf8(l)))
    if b"testament3-sha1" in metadata.verifiers:
        lines.append(b"testament3-sha1: %s\n" %
                     metadata.verifiers[b"testament3-sha1"])
    return b"".join(lines)


def extract_bzr_metadata(message):
    """Extract Bazaar metadata from a commit message.

    :param message: Commit message to extract from
    :return: Tuple with original commit message and metadata object
    """
    split = message.split(b"\n--BZR--\n", 1)
    if len(split) != 2:
        return message, None
    return split[0], parse_roundtripping_metadata(split[1])


def inject_bzr_metadata(message, commit_supplement, encoding):
    if not commit_supplement:
        return message
    rt_data = generate_roundtripping_metadata(commit_supplement, encoding)
    if not rt_data:
        return message
    if not isinstance(rt_data, bytes):
        raise TypeError(rt_data)
    return message + b"\n--BZR--\n" + rt_data



# Copyright (C) 2006, 2007, 2009-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for display of exceptions."""

import os
import re
import unittest

from breezy import bzr, config, controldir, errors, osutils, repository, tests
from breezy.bzr.groupcompress_repo import RepositoryFormat2a


class TestExceptionReporting(tests.TestCaseInTempDir):

    def test_exception_exitcode(self):
        # we must use a subprocess, because the normal in-memory mechanism
        # allows errors to propagate up through the test suite
        out, err = self.run_brz_subprocess(['assert-fail'],
                                           universal_newlines=True,
                                           retcode=errors.EXIT_INTERNAL_ERROR)
        self.assertEqual(4, errors.EXIT_INTERNAL_ERROR)
        self.assertContainsRe(err, br'\nAssertionError: always fails\n')
        self.assertContainsRe(
            err, br'Breezy has encountered an internal error')

    @unittest.skip("encoding when LANG=C is currently borked")
    def test_utf8_default_fs_enc(self):
        """In the C locale brz treats a posix filesystem as UTF-8 encoded"""
        if os.name != "posix":
            raise tests.TestNotApplicable("Needs system beholden to C locales")
        out, err = self.run_brz_subprocess(["init", "file:%C2%A7"],
                                           env_changes={"LANG": "C", "LC_ALL": "C"})
        self.assertContainsRe(out, b"^Created a standalone tree .*$")


class TestOptParseBugHandling(tests.TestCase):
    "Test that we handle http://bugs.python.org/issue2931"

    def test_nonascii_optparse(self):
        """Reasonable error raised when non-ascii in option name on Python 2"""
        error_re = 'no such option: -\xe4'
        out = self.run_bzr_error([error_re], ['st', '-\xe4'])


class TestObsoleteRepoFormat(RepositoryFormat2a):

    @classmethod
    def get_format_string(cls):
        return b"Test Obsolete Repository Format"

    def is_deprecated(self):
        return True


class TestDeprecationWarning(tests.TestCaseWithTransport):
    """The deprecation warning is controlled via a global variable:
    repository._deprecation_warning_done. As such, it can be emitted only once
    during a brz invocation, no matter how many repositories are involved.

    It would be better if it was a repo attribute instead but that's far more
    work than I want to do right now -- vila 20091215.
    """

    def setUp(self):
        super().setUp()
        self.addCleanup(repository.format_registry.remove,
                        TestObsoleteRepoFormat)
        repository.format_registry.register(TestObsoleteRepoFormat)
        self.addCleanup(controldir.format_registry.remove, "testobsolete")
        bzr.register_metadir(controldir.format_registry, "testobsolete",
                             "breezy.tests.blackbox.test_exceptions.TestObsoleteRepoFormat",
                             branch_format='breezy.bzr.branch.BzrBranchFormat7',
                             tree_format='breezy.bzr.workingtree_4.WorkingTreeFormat6',
                             deprecated=True,
                             help='Same as 2a, but with an obsolete repo format.')
        self.disable_deprecation_warning()

    def enable_deprecation_warning(self, repo=None):
        """repo is not used yet since _deprecation_warning_done is a global"""
        repository._deprecation_warning_done = False

    def disable_deprecation_warning(self, repo=None):
        """repo is not used yet since _deprecation_warning_done is a global"""
        repository._deprecation_warning_done = True

    def make_obsolete_repo(self, path):
        # We don't want the deprecation raising during the repo creation
        format = controldir.format_registry.make_controldir("testobsolete")
        tree = self.make_branch_and_tree(path, format=format)
        return tree

    def check_warning(self, present):
        if present:
            check = self.assertContainsRe
        else:
            check = self.assertNotContainsRe
        check(self.get_log(), 'WARNING.*brz upgrade')

    def test_repository_deprecation_warning(self):
        """Old formats give a warning"""
        self.make_obsolete_repo('foo')
        self.enable_deprecation_warning()
        out, err = self.run_bzr('status', working_dir='foo')
        self.check_warning(True)

    def test_repository_deprecation_warning_suppressed_global(self):
        """Old formats give a warning"""
        conf = config.GlobalStack()
        conf.set('suppress_warnings', 'format_deprecation')
        self.make_obsolete_repo('foo')
        self.enable_deprecation_warning()
        out, err = self.run_bzr('status', working_dir='foo')
        self.check_warning(False)

    def test_repository_deprecation_warning_suppressed_locations(self):
        """Old formats give a warning"""
        self.make_obsolete_repo('foo')
        conf = config.LocationStack(osutils.pathjoin(self.test_dir, 'foo'))
        conf.set('suppress_warnings', 'format_deprecation')
        self.enable_deprecation_warning()
        out, err = self.run_bzr('status', working_dir='foo')
        self.check_warning(False)

    def test_repository_deprecation_warning_suppressed_branch(self):
        """Old formats give a warning"""
        tree = self.make_obsolete_repo('foo')
        conf = tree.branch.get_config_stack()
        conf.set('suppress_warnings', 'format_deprecation')
        self.enable_deprecation_warning()
        out, err = self.run_bzr('status', working_dir='foo')
        self.check_warning(False)



# Copyright (C) 2010, 2011, 2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Black-box tests for brz config."""

from ... import config, tests
from .. import script
from .. import test_config as _t_config


class TestWithoutConfig(tests.TestCaseWithTransport):

    def test_config_all(self):
        out, err = self.run_bzr(['config'])
        self.assertEqual('', out)
        self.assertEqual('', err)

    def test_remove_unknown_option(self):
        self.run_bzr_error(['The "file" configuration option does not exist', ],
                           ['config', '--remove', 'file'])

    def test_all_remove_exclusive(self):
        self.run_bzr_error(['--all and --remove are mutually exclusive.', ],
                           ['config', '--remove', '--all'])

    def test_all_set_exclusive(self):
        self.run_bzr_error(['Only one option can be set.', ],
                           ['config', '--all', 'hello=world'])

    def test_remove_no_option(self):
        self.run_bzr_error(['--remove expects an option to remove.', ],
                           ['config', '--remove'])

    def test_unknown_option(self):
        self.run_bzr_error(['The "file" configuration option does not exist', ],
                           ['config', 'file'])

    def test_unexpected_regexp(self):
        self.run_bzr_error(
            ['The "\\*file" configuration option does not exist', ],
            ['config', '*file'])

    def test_wrong_regexp(self):
        self.run_bzr_error(
            ['Invalid pattern\\(s\\) found. "\\*file" nothing to repeat', ],
            ['config', '--all', '*file'])


class TestConfigDisplay(tests.TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        _t_config.create_configs(self)

    def test_multiline_all_values(self):
        self.breezy_config.set_user_option('multiline', '1\n2\n')
        # Fallout from bug 710410, the triple quotes have been toggled
        script.run_script(self, '''\
            $ brz config -d tree
            breezy:
              [DEFAULT]
              multiline = """1
            2
            """
            ''')

    def test_multiline_value_only(self):
        self.breezy_config.set_user_option('multiline', '1\n2\n')
        # Fallout from bug 710410, the triple quotes have been toggled
        script.run_script(self, '''\
            $ brz config -d tree multiline
            """1
            2
            """
            ''')

    def test_list_value_all(self):
        config.option_registry.register(config.ListOption('list'))
        self.addCleanup(config.option_registry.remove, 'list')
        self.breezy_config.set_user_option('list', [1, 'a', 'with, a comma'])
        script.run_script(self, '''\
            $ brz config -d tree
            breezy:
              [DEFAULT]
              list = 1, a, "with, a comma"
            ''')

    def test_list_value_one(self):
        config.option_registry.register(config.ListOption('list'))
        self.addCleanup(config.option_registry.remove, 'list')
        self.breezy_config.set_user_option('list', [1, 'a', 'with, a comma'])
        script.run_script(self, '''\
            $ brz config -d tree list
            1, a, "with, a comma"
            ''')

    def test_registry_value_all(self):
        self.breezy_config.set_user_option('transform.orphan_policy',
                                           'move')
        script.run_script(self, '''\
            $ brz config -d tree
            breezy:
              [DEFAULT]
              transform.orphan_policy = move
            ''')

    def test_registry_value_one(self):
        self.breezy_config.set_user_option('transform.orphan_policy',
                                           'move')
        script.run_script(self, '''\
            $ brz config -d tree transform.orphan_policy
            move
            ''')

    def test_breezy_config(self):
        self.breezy_config.set_user_option('hello', 'world')
        script.run_script(self, '''\
            $ brz config -d tree
            breezy:
              [DEFAULT]
              hello = world
            ''')

    def test_locations_config_for_branch(self):
        self.locations_config.set_user_option('hello', 'world')
        self.branch_config.set_user_option('hello', 'you')
        script.run_script(self, '''\
            $ brz config -d tree
            locations:
              [.../tree]
              hello = world
            branch:
              hello = you
            ''')

    def test_locations_config_outside_branch(self):
        self.breezy_config.set_user_option('hello', 'world')
        self.locations_config.set_user_option('hello', 'world')
        script.run_script(self, '''\
            $ brz config
            breezy:
              [DEFAULT]
              hello = world
            ''')

    def test_cmd_line(self):
        self.breezy_config.set_user_option('hello', 'world')
        script.run_script(self, '''\
            $ brz config -Ohello=bzr
            cmdline:
              hello = bzr
            breezy:
              [DEFAULT]
              hello = world
            ''')


class TestConfigDisplayWithPolicy(tests.TestCaseWithTransport):

    def test_location_with_policy(self):
        # LocationConfig is the only one dealing with policies so far.
        self.make_branch_and_tree('tree')
        config_text = """\
[{dir}]
url = dir
url:policy = appendpath
[{dir}/tree]
url = tree
""".format(dir=self.test_dir)
        # We don't use the config directly so we save it to disk
        config.LocationConfig.from_string(config_text, 'tree', save=True)
        # policies are displayed with their options since they are part of
        # their definition, likewise the path is not appended, we are just
        # presenting the relevant portions of the config files
        script.run_script(self, '''\
            $ brz config -d tree --all url
            locations:
              [.../work/tree]
              url = tree
              [.../work]
              url = dir
              url:policy = appendpath
            ''')


class TestConfigActive(tests.TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        _t_config.create_configs_with_file_option(self)

    def test_active_in_locations(self):
        script.run_script(self, '''\
            $ brz config -d tree file
            locations
            ''')

    def test_active_in_breezy(self):
        script.run_script(self, '''\
            $ brz config -d tree --scope breezy file
            breezy
            ''')

    def test_active_in_branch(self):
        # We need to delete the locations definition that overrides the branch
        # one
        script.run_script(self, '''\
            $ brz config -d tree --scope locations --remove file
            $ brz config -d tree file
            branch
            ''')


class TestConfigSetOption(tests.TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        _t_config.create_configs(self)

    def test_unknown_config(self):
        self.run_bzr_error(['The "moon" configuration does not exist'],
                           ['config', '--scope', 'moon', 'hello=world'])

    def test_breezy_config_outside_branch(self):
        script.run_script(self, '''\
            $ brz config --scope breezy hello=world
            $ brz config -d tree --all hello
            breezy:
              [DEFAULT]
              hello = world
            ''')

    def test_breezy_config_inside_branch(self):
        script.run_script(self, '''\
            $ brz config -d tree --scope breezy hello=world
            $ brz config -d tree --all hello
            breezy:
              [DEFAULT]
              hello = world
            ''')

    def test_locations_config_inside_branch(self):
        script.run_script(self, '''\
            $ brz config -d tree --scope locations hello=world
            $ brz config -d tree --all hello
            locations:
              [.../work/tree]
              hello = world
            ''')

    def test_branch_config_default(self):
        script.run_script(self, '''\
            $ brz config -d tree hello=world
            $ brz config -d tree --all hello
            branch:
              hello = world
            ''')

    def test_branch_config_forcing_branch(self):
        script.run_script(self, '''\
            $ brz config -d tree --scope branch hello=world
            $ brz config -d tree --all hello
            branch:
              hello = world
            ''')


class TestConfigRemoveOption(tests.TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        _t_config.create_configs_with_file_option(self)

    def test_unknown_config(self):
        self.run_bzr_error(['The "moon" configuration does not exist'],
                           ['config', '--scope', 'moon', '--remove', 'file'])

    def test_breezy_config_outside_branch(self):
        script.run_script(self, '''\
            $ brz config --scope breezy --remove file
            $ brz config -d tree --all file
            locations:
              [.../work/tree]
              file = locations
            branch:
              file = branch
            ''')

    def test_breezy_config_inside_branch(self):
        script.run_script(self, '''\
            $ brz config -d tree --scope breezy --remove file
            $ brz config -d tree --all file
            locations:
              [.../work/tree]
              file = locations
            branch:
              file = branch
            ''')

    def test_locations_config_inside_branch(self):
        script.run_script(self, '''\
            $ brz config -d tree --scope locations --remove file
            $ brz config -d tree --all file
            branch:
              file = branch
            breezy:
              [DEFAULT]
              file = breezy
            ''')

    def test_branch_config_default(self):
        script.run_script(self, '''\
            $ brz config -d tree --scope locations --remove file
            $ brz config -d tree --all file
            branch:
              file = branch
            breezy:
              [DEFAULT]
              file = breezy
            ''')
        script.run_script(self, '''\
            $ brz config -d tree --remove file
            $ brz config -d tree --all file
            breezy:
              [DEFAULT]
              file = breezy
            ''')

    def test_branch_config_forcing_branch(self):
        script.run_script(self, '''\
            $ brz config -d tree --scope branch --remove file
            $ brz config -d tree --all file
            locations:
              [.../work/tree]
              file = locations
            breezy:
              [DEFAULT]
              file = breezy
            ''')
        script.run_script(self, '''\
            $ brz config -d tree --scope locations --remove file
            $ brz config -d tree --all file
            breezy:
              [DEFAULT]
              file = breezy
            ''')


class TestConfigDirectory(tests.TestCaseWithTransport):

    def test_parent_alias(self):
        t = self.make_branch_and_tree('base')
        t.branch.get_config_stack().set('test', 'base')
        clone = t.branch.controldir.sprout('clone').open_branch()
        clone.get_config_stack().set('test', 'clone')
        out, err = self.run_bzr(['config', '-d', ':parent', 'test'],
                                working_dir='clone')
        self.assertEqual('base\n', out)



# Copyright (C) 2005, 2006, 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Transport implementation that adapts another transport to look like NFS.

Currently this means that the rename() call will raise ResourceBusy when a
target path is a directory.

To get a fake nfs transport use get_transport('fakenfs+' + real_url)
"""

from stat import S_ISDIR

from .. import errors
from .. import transport as _mod_transport
from .. import urlutils
from . import decorator


class FakeNFSTransportDecorator(decorator.TransportDecorator):
    """A transport that behaves like NFS, for testing"""

    @classmethod
    def _get_url_prefix(self):
        """FakeNFS transports are identified by 'fakenfs+'"""
        return 'fakenfs+'

    def rename(self, rel_from, rel_to):
        """See Transport.rename().

        This variation on rename converts DirectoryNotEmpty and FileExists
        errors into ResourceBusy if the target is a directory.
        """
        try:
            self._decorated.rename(rel_from, rel_to)
        except (errors.DirectoryNotEmpty, _mod_transport.FileExists) as e:
            # if this is a directory rename, raise
            # resourcebusy rather than DirectoryNotEmpty
            stat = self._decorated.stat(rel_to)
            if S_ISDIR(stat.st_mode):
                raise errors.ResourceBusy(rel_to)
            else:
                raise

    def delete(self, relpath):
        if urlutils.basename(relpath).startswith('.nfs'):
            raise errors.ResourceBusy(self.abspath(relpath))
        return self._decorated.delete(relpath)


def get_test_permutations():
    """Return the permutations to be used in testing."""
    from breezy.tests import test_server
    return [(FakeNFSTransportDecorator, test_server.FakeNFSServer), ]



# Copyright (C) 2008-2012 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import os
import sys

from .... import (bedding, controldir, errors, osutils, revisionspec, tests,
                  transport, uncommit, workingtree)
from ....tests import features, per_branch, per_transport
from .. import cmds


def get_transport_scenarios():
    result = []
    basis = per_transport.transport_test_permutations()
    # Keep only the interesting ones for upload
    usable_classes = set()
    if features.paramiko.available():
        from ....transport import sftp
        usable_classes.add(sftp.SFTPTransport)
    from ....transport import local
    usable_classes.add(local.LocalTransport)
    for name, d in basis:
        t_class = d['transport_class']
        if t_class in usable_classes:
            result.append((name, d))
    return result


def load_tests(loader, standard_tests, pattern):
    """Multiply tests for tranport implementations."""
    result = loader.suiteClass()

    # one for each transport implementation
    t_tests, remaining_tests = tests.split_suite_by_condition(
        standard_tests, tests.condition_isinstance((
            TestFullUpload,
            TestIncrementalUpload,
            TestUploadFromRemoteBranch,
            )))
    tests.multiply_tests(t_tests, get_transport_scenarios(), result)

    # one for each branch format
    b_tests, remaining_tests = tests.split_suite_by_condition(
        remaining_tests, tests.condition_isinstance((
            TestBranchUploadLocations,
            )))
    tests.multiply_tests(b_tests, per_branch.branch_scenarios(),
                         result)

    # No parametrization for the remaining tests
    result.addTests(remaining_tests)

    return result


class UploadUtilsMixin:
    """Helper class to write upload tests.

    This class provides helpers to simplify test writing. The emphasis is on
    easy test writing, so each tree modification is committed. This doesn't
    preclude writing tests spawning several revisions to upload more complex
    changes.
    """

    upload_dir = 'upload'
    branch_dir = 'branch'

    def make_branch_and_working_tree(self):
        t = transport.get_transport(self.branch_dir)
        t.ensure_base()
        branch = controldir.ControlDir.create_branch_convenience(
            t.base,
            format=controldir.format_registry.make_controldir('default'),
            force_new_tree=False)
        self.tree = branch.controldir.create_workingtree()
        self.tree.commit('initial empty tree')

    def assertUpFileEqual(self, content, path, base=upload_dir):
        self.assertFileEqual(content, osutils.pathjoin(base, path))

    def assertUpPathModeEqual(self, path, expected_mode, base=upload_dir):
        # FIXME: the tests needing that assertion should depend on the server
        # ability to handle chmod so that they don't fail (or be skipped)
        # against servers that can't. Note that some breezy transports define
        # _can_roundtrip_unix_modebits in a incomplete way, this property
        # should depend on both the client and the server, not the client only.
        # But the client will know or can find if the server support chmod so
        # that's the client that will report it anyway.
        full_path = osutils.pathjoin(base, path)
        st = os.stat(full_path)
        mode = st.st_mode & 0o777
        if expected_mode == mode:
            return
        raise AssertionError(
            'For path %s, mode is %s not %s' %
            (full_path, oct(mode), oct(expected_mode)))

    def assertUpPathDoesNotExist(self, path, base=upload_dir):
        self.assertPathDoesNotExist(osutils.pathjoin(base, path))

    def assertUpPathExists(self, path, base=upload_dir):
        self.assertPathExists(osutils.pathjoin(base, path))

    def set_file_content(self, path, content, base=branch_dir):
        with open(osutils.pathjoin(base, path), 'wb') as f:
            f.write(content)

    def add_file(self, path, content, base=branch_dir):
        self.set_file_content(path, content, base)
        self.tree.add(path)
        self.tree.commit('add file %s' % path)

    def modify_file(self, path, content, base=branch_dir):
        self.set_file_content(path, content, base)
        self.tree.commit('modify file %s' % path)

    def chmod_file(self, path, mode, base=branch_dir):
        full_path = osutils.pathjoin(base, path)
        os.chmod(full_path, mode)
        self.tree.commit('change file {} mode to {}'.format(path, oct(mode)))

    def delete_any(self, path, base=branch_dir):
        self.tree.remove([path], keep_files=False)
        self.tree.commit('delete %s' % path)

    def add_dir(self, path, base=branch_dir):
        os.mkdir(osutils.pathjoin(base, path))
        self.tree.add(path)
        self.tree.commit('add directory %s' % path)

    def rename_any(self, old_path, new_path):
        self.tree.rename_one(old_path, new_path)
        self.tree.commit('rename {} into {}'.format(old_path, new_path))

    def transform_dir_into_file(self, path, content, base=branch_dir):
        osutils.delete_any(osutils.pathjoin(base, path))
        self.set_file_content(path, content, base)
        self.tree.commit('change %s from dir to file' % path)

    def transform_file_into_dir(self, path, base=branch_dir):
        # bzr can't handle that kind change in a single commit without an
        # intervening bzr status (see bug #205636).
        self.tree.remove([path], keep_files=False)
        os.mkdir(osutils.pathjoin(base, path))
        self.tree.add(path)
        self.tree.commit('change %s from file to dir' % path)

    def add_symlink(self, path, target, base=branch_dir):
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        os.symlink(target, osutils.pathjoin(base, path))
        self.tree.add(path)
        self.tree.commit('add symlink {} -> {}'.format(path, target))

    def modify_symlink(self, path, target, base=branch_dir):
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        full_path = osutils.pathjoin(base, path)
        os.unlink(full_path)
        os.symlink(target, full_path)
        self.tree.commit('modify symlink {} -> {}'.format(path, target))

    def _get_cmd_upload(self):
        cmd = cmds.cmd_upload()
        # We don't want to use run_bzr here because redirected output are a
        # pain to debug. But we need to provides a valid outf.
        # XXX: Should a bug against bzr be filled about that ?

        # Short story: we don't expect any output so we may just use stdout
        cmd.outf = sys.stdout
        return cmd

    def do_full_upload(self, *args, **kwargs):
        upload = self._get_cmd_upload()
        up_url = self.get_url(self.upload_dir)
        if kwargs.get('directory', None) is None:
            kwargs['directory'] = self.branch_dir
        kwargs['full'] = True
        kwargs['quiet'] = True
        upload.run(up_url, *args, **kwargs)

    def do_incremental_upload(self, *args, **kwargs):
        upload = self._get_cmd_upload()
        up_url = self.get_url(self.upload_dir)
        if kwargs.get('directory', None) is None:
            kwargs['directory'] = self.branch_dir
        kwargs['quiet'] = True
        upload.run(up_url, *args, **kwargs)


class TestUploadMixin(UploadUtilsMixin):
    """Helper class to share tests between full and incremental uploads."""

    def _test_create_file(self, file_name):
        self.make_branch_and_working_tree()
        self.do_full_upload()
        self.add_file(file_name, b'foo')

        self.do_upload()

        self.assertUpFileEqual(b'foo', file_name)

    def test_create_file(self):
        self._test_create_file('hello')

    def test_unicode_create_file(self):
        self.requireFeature(features.UnicodeFilenameFeature)
        self._test_create_file('hell\u00d8')

    def _test_create_file_in_dir(self, dir_name, file_name):
        self.make_branch_and_working_tree()
        self.do_full_upload()
        self.add_dir(dir_name)
        fpath = '{}/{}'.format(dir_name, file_name)
        self.add_file(fpath, b'baz')

        self.assertUpPathDoesNotExist(fpath)

        self.do_upload()

        self.assertUpFileEqual(b'baz', fpath)
        self.assertUpPathModeEqual(dir_name, 0o775)

    def test_create_file_in_dir(self):
        self._test_create_file_in_dir('dir', 'goodbye')

    def test_unicode_create_file_in_dir(self):
        self.requireFeature(features.UnicodeFilenameFeature)
        self._test_create_file_in_dir('dir\u00d8', 'goodbye\u00d8')

    def test_modify_file(self):
        self.make_branch_and_working_tree()
        self.add_file('hello', b'foo')
        self.do_full_upload()
        self.modify_file('hello', b'bar')

        self.assertUpFileEqual(b'foo', 'hello')

        self.do_upload()

        self.assertUpFileEqual(b'bar', 'hello')

    def _test_rename_one_file(self, old_name, new_name):
        self.make_branch_and_working_tree()
        self.add_file(old_name, b'foo')
        self.do_full_upload()
        self.rename_any(old_name, new_name)

        self.assertUpFileEqual(b'foo', old_name)

        self.do_upload()

        self.assertUpFileEqual(b'foo', new_name)

    def test_rename_one_file(self):
        self._test_rename_one_file('hello', 'goodbye')

    def test_unicode_rename_one_file(self):
        self.requireFeature(features.UnicodeFilenameFeature)
        self._test_rename_one_file('hello\u00d8', 'goodbye\u00d8')

    def test_rename_and_change_file(self):
        self.make_branch_and_working_tree()
        self.add_file('hello', b'foo')
        self.do_full_upload()
        self.rename_any('hello', 'goodbye')
        self.modify_file('goodbye', b'bar')

        self.assertUpFileEqual(b'foo', 'hello')

        self.do_upload()

        self.assertUpFileEqual(b'bar', 'goodbye')

    def test_rename_two_files(self):
        self.make_branch_and_working_tree()
        self.add_file('a', b'foo')
        self.add_file('b', b'qux')
        self.do_full_upload()
        # We rely on the assumption that bzr will topologically sort the
        # renames which will cause a -> b to appear *before* b -> c
        self.rename_any('b', 'c')
        self.rename_any('a', 'b')

        self.assertUpFileEqual(b'foo', 'a')
        self.assertUpFileEqual(b'qux', 'b')

        self.do_upload()

        self.assertUpFileEqual(b'foo', 'b')
        self.assertUpFileEqual(b'qux', 'c')

    def test_upload_revision(self):
        self.make_branch_and_working_tree()  # rev1
        self.do_full_upload()
        self.add_file('hello', b'foo')  # rev2
        self.modify_file('hello', b'bar')  # rev3

        self.assertUpPathDoesNotExist('hello')

        revspec = revisionspec.RevisionSpec.from_string('2')
        self.do_upload(revision=[revspec])

        self.assertUpFileEqual(b'foo', 'hello')

    def test_no_upload_when_changes(self):
        self.make_branch_and_working_tree()
        self.add_file('a', b'foo')
        self.set_file_content('a', b'bar')

        self.assertRaises(errors.UncommittedChanges, self.do_upload)

    def test_no_upload_when_conflicts(self):
        self.make_branch_and_working_tree()
        self.add_file('a', b'foo')
        self.run_bzr('branch branch other')
        self.modify_file('a', b'bar')
        other_tree = workingtree.WorkingTree.open('other')
        self.set_file_content('a', b'baz', 'other/')
        other_tree.commit('modify file a')

        self.run_bzr('merge -d branch other', retcode=1)

        self.assertRaises(errors.UncommittedChanges, self.do_upload)

    def _test_change_file_into_dir(self, file_name):
        self.make_branch_and_working_tree()
        self.add_file(file_name, b'foo')
        self.do_full_upload()
        self.transform_file_into_dir(file_name)
        fpath = '{}/{}'.format(file_name, 'file')
        self.add_file(fpath, b'bar')

        self.assertUpFileEqual(b'foo', file_name)

        self.do_upload()

        self.assertUpFileEqual(b'bar', fpath)

    def test_change_file_into_dir(self):
        self._test_change_file_into_dir('hello')

    def test_unicode_change_file_into_dir(self):
        self.requireFeature(features.UnicodeFilenameFeature)
        self._test_change_file_into_dir('hello\u00d8')

    def test_change_dir_into_file(self):
        self.make_branch_and_working_tree()
        self.add_dir('hello')
        self.add_file('hello/file', b'foo')
        self.do_full_upload()
        self.delete_any('hello/file')
        self.transform_dir_into_file('hello', b'bar')

        self.assertUpFileEqual(b'foo', 'hello/file')

        self.do_upload()

        self.assertUpFileEqual(b'bar', 'hello')

    def _test_make_file_executable(self, file_name):
        self.make_branch_and_working_tree()
        self.add_file(file_name, b'foo')
        self.chmod_file(file_name, 0o664)
        self.do_full_upload()
        self.chmod_file(file_name, 0o755)

        self.assertUpPathModeEqual(file_name, 0o664)

        self.do_upload()

        self.assertUpPathModeEqual(file_name, 0o775)

    def test_make_file_executable(self):
        self._test_make_file_executable('hello')

    def test_unicode_make_file_executable(self):
        self.requireFeature(features.UnicodeFilenameFeature)
        self._test_make_file_executable('hello\u00d8')

    def test_create_symlink(self):
        self.make_branch_and_working_tree()
        self.do_full_upload()
        self.add_symlink('link', 'target')

        self.do_upload()

        self.assertUpPathExists('link')

    def test_rename_symlink(self):
        self.make_branch_and_working_tree()
        old_name, new_name = 'old-link', 'new-link'
        self.add_symlink(old_name, 'target')
        self.do_full_upload()

        self.rename_any(old_name, new_name)

        self.do_upload()

        self.assertUpPathExists(new_name)

    def get_upload_auto(self):
        # We need a fresh branch to check what has been saved on disk
        b = controldir.ControlDir.open(self.tree.basedir).open_branch()
        return b.get_config_stack().get('upload_auto')

    def test_upload_auto(self):
        """Test that upload --auto sets the upload_auto option"""
        self.make_branch_and_working_tree()

        self.add_file('hello', b'foo')
        self.assertFalse(self.get_upload_auto())
        self.do_full_upload(auto=True)
        self.assertUpFileEqual(b'foo', 'hello')
        self.assertTrue(self.get_upload_auto())

        # and check that it stays set until it is unset
        self.add_file('bye', b'bar')
        self.do_full_upload()
        self.assertUpFileEqual(b'bar', 'bye')
        self.assertTrue(self.get_upload_auto())

    def test_upload_noauto(self):
        """Test that upload --no-auto unsets the upload_auto option"""
        self.make_branch_and_working_tree()

        self.add_file('hello', b'foo')
        self.do_full_upload(auto=True)
        self.assertUpFileEqual(b'foo', 'hello')
        self.assertTrue(self.get_upload_auto())

        self.add_file('bye', b'bar')
        self.do_full_upload(auto=False)
        self.assertUpFileEqual(b'bar', 'bye')
        self.assertFalse(self.get_upload_auto())

        # and check that it stays unset until it is set
        self.add_file('again', b'baz')
        self.do_full_upload()
        self.assertUpFileEqual(b'baz', 'again')
        self.assertFalse(self.get_upload_auto())

    def test_upload_from_subdir(self):
        self.make_branch_and_working_tree()
        self.build_tree(['branch/foo/', 'branch/foo/bar'])
        self.tree.add(['foo/', 'foo/bar'])
        self.tree.commit("Add directory")
        self.do_full_upload(directory='branch/foo')

    def test_upload_revid_path_in_dir(self):
        self.make_branch_and_working_tree()
        self.add_dir('dir')
        self.add_file('dir/goodbye', b'baz')

        revid_path = 'dir/revid-path'
        self.tree.branch.get_config_stack(
            ).set('upload_revid_location', revid_path)
        self.assertUpPathDoesNotExist(revid_path)

        self.do_full_upload()

        self.add_file('dir/hello', b'foo')

        self.do_upload()

        self.assertUpPathExists(revid_path)
        self.assertUpFileEqual(b'baz', 'dir/goodbye')
        self.assertUpFileEqual(b'foo', 'dir/hello')

    def test_ignore_file(self):
        self.make_branch_and_working_tree()
        self.do_full_upload()
        self.add_file('.bzrignore-upload', b'foo')
        self.add_file('foo', b'bar')

        self.do_upload()

        self.assertUpPathDoesNotExist('foo')

    def test_ignore_regexp(self):
        self.make_branch_and_working_tree()
        self.do_full_upload()
        self.add_file('.bzrignore-upload', b'f*')
        self.add_file('foo', b'bar')

        self.do_upload()

        self.assertUpPathDoesNotExist('foo')

    def test_ignore_directory(self):
        self.make_branch_and_working_tree()
        self.do_full_upload()
        self.add_file('.bzrignore-upload', b'dir')
        self.add_dir('dir')

        self.do_upload()

        self.assertUpPathDoesNotExist('dir')

    def test_ignore_nested_directory(self):
        self.make_branch_and_working_tree()
        self.do_full_upload()
        self.add_file('.bzrignore-upload', b'dir')
        self.add_dir('dir')
        self.add_dir('dir/foo')
        self.add_file('dir/foo/bar', b'bar contents')

        self.do_upload()

        self.assertUpPathDoesNotExist('dir')
        self.assertUpPathDoesNotExist('dir/foo/bar')

    def test_ignore_change_file_into_dir(self):
        self.make_branch_and_working_tree()
        self.add_file('hello', b'foo')
        self.do_full_upload()
        self.add_file('.bzrignore-upload', b'hello')
        self.transform_file_into_dir('hello')
        self.add_file('hello/file', b'bar')

        self.assertUpFileEqual(b'foo', 'hello')

        self.do_upload()

        self.assertUpFileEqual(b'foo', 'hello')

    def test_ignore_change_dir_into_file(self):
        self.make_branch_and_working_tree()
        self.add_dir('hello')
        self.add_file('hello/file', b'foo')
        self.do_full_upload()

        self.add_file('.bzrignore-upload', b'hello')
        self.delete_any('hello/file')
        self.transform_dir_into_file('hello', b'bar')

        self.assertUpFileEqual(b'foo', 'hello/file')

        self.do_upload()

        self.assertUpFileEqual(b'foo', 'hello/file')

    def test_ignore_delete_dir_in_subdir(self):
        self.make_branch_and_working_tree()
        self.add_dir('dir')
        self.add_dir('dir/subdir')
        self.add_file('dir/subdir/a', b'foo')
        self.do_full_upload()
        self.add_file('.bzrignore-upload', b'dir/subdir')
        self.rename_any('dir/subdir/a', 'dir/a')
        self.delete_any('dir/subdir')

        self.assertUpFileEqual(b'foo', 'dir/subdir/a')

        self.do_upload()

        # The file in the dir is not ignored. This a bit contrived but
        # indicates that we may encounter problems when ignored items appear
        # and disappear... -- vila 100106
        self.assertUpFileEqual(b'foo', 'dir/a')


class TestFullUpload(tests.TestCaseWithTransport, TestUploadMixin):

    do_upload = TestUploadMixin.do_full_upload

    def test_full_upload_empty_tree(self):
        self.make_branch_and_working_tree()

        self.do_full_upload()

        revid_path = self.tree.branch.get_config_stack(
            ).get('upload_revid_location')
        self.assertUpPathExists(revid_path)

    def test_invalid_revspec(self):
        self.make_branch_and_working_tree()
        rev1 = revisionspec.RevisionSpec.from_string('1')
        rev2 = revisionspec.RevisionSpec.from_string('2')

        self.assertRaises(errors.CommandError,
                          self.do_incremental_upload, revision=[rev1, rev2])

    def test_create_remote_dir_twice(self):
        self.make_branch_and_working_tree()
        self.add_dir('dir')
        self.do_full_upload()
        self.add_file('dir/goodbye', b'baz')

        self.assertUpPathDoesNotExist('dir/goodbye')

        self.do_full_upload()

        self.assertUpFileEqual(b'baz', 'dir/goodbye')
        self.assertUpPathModeEqual('dir', 0o775)


class TestIncrementalUpload(tests.TestCaseWithTransport, TestUploadMixin):

    do_upload = TestUploadMixin.do_incremental_upload

    # XXX: full upload doesn't handle deletions....

    def test_delete_one_file(self):
        self.make_branch_and_working_tree()
        self.add_file('hello', b'foo')
        self.do_full_upload()
        self.delete_any('hello')

        self.assertUpFileEqual(b'foo', 'hello')

        self.do_upload()

        self.assertUpPathDoesNotExist('hello')

    def test_delete_dir_and_subdir(self):
        self.make_branch_and_working_tree()
        self.add_dir('dir')
        self.add_dir('dir/subdir')
        self.add_file('dir/subdir/a', b'foo')
        self.do_full_upload()
        self.rename_any('dir/subdir/a', 'a')
        self.delete_any('dir/subdir')
        self.delete_any('dir')

        self.assertUpFileEqual(b'foo', 'dir/subdir/a')

        self.do_upload()

        self.assertUpPathDoesNotExist('dir/subdir/a')
        self.assertUpPathDoesNotExist('dir/subdir')
        self.assertUpPathDoesNotExist('dir')
        self.assertUpFileEqual(b'foo', 'a')

    def test_delete_one_file_rename_to_deleted(self):
        self.make_branch_and_working_tree()
        self.add_file('a', b'foo')
        self.add_file('b', b'bar')
        self.do_full_upload()
        self.delete_any('a')
        self.rename_any('b', 'a')

        self.assertUpFileEqual(b'foo', 'a')

        self.do_upload()

        self.assertUpPathDoesNotExist('b')
        self.assertUpFileEqual(b'bar', 'a')

    def test_rename_outside_dir_delete_dir(self):
        self.make_branch_and_working_tree()
        self.add_dir('dir')
        self.add_file('dir/a', b'foo')
        self.do_full_upload()
        self.rename_any('dir/a', 'a')
        self.delete_any('dir')

        self.assertUpFileEqual(b'foo', 'dir/a')

        self.do_upload()

        self.assertUpPathDoesNotExist('dir/a')
        self.assertUpPathDoesNotExist('dir')
        self.assertUpFileEqual(b'foo', 'a')

    def test_delete_symlink(self):
        self.make_branch_and_working_tree()
        self.add_symlink('link', 'target')
        self.do_full_upload()
        self.delete_any('link')

        self.do_upload()

        self.assertUpPathDoesNotExist('link')

    def test_upload_for_the_first_time_do_a_full_upload(self):
        self.make_branch_and_working_tree()
        self.add_file('hello', b'bar')

        revid_path = self.tree.branch.get_config_stack(
            ).get('upload_revid_location')
        self.assertUpPathDoesNotExist(revid_path)

        self.do_upload()

        self.assertUpFileEqual(b'bar', 'hello')

    def test_ignore_delete_one_file(self):
        self.make_branch_and_working_tree()
        self.add_file('hello', b'foo')
        self.do_full_upload()
        self.add_file('.bzrignore-upload', b'hello')
        self.delete_any('hello')

        self.assertUpFileEqual(b'foo', 'hello')

        self.do_upload()

        self.assertUpFileEqual(b'foo', 'hello')


class TestBranchUploadLocations(per_branch.TestCaseWithBranch):

    def test_get_upload_location_unset(self):
        conf = self.get_branch().get_config_stack()
        self.assertEqual(None, conf.get('upload_location'))

    def test_get_push_location_exact(self):
        bedding.ensure_config_dir_exists()
        fn = bedding.locations_config_path()
        b = self.get_branch()
        with open(fn, 'w') as f:
            f.write("[%s]\n" "upload_location=foo\n" % b.base.rstrip("/"))
        self.assertEqual("foo", b.get_config_stack().get('upload_location'))

    def test_set_push_location(self):
        conf = self.get_branch().get_config_stack()
        conf.set('upload_location', 'foo')
        self.assertEqual('foo', conf.get('upload_location'))


class TestUploadFromRemoteBranch(tests.TestCaseWithTransport, UploadUtilsMixin):

    remote_branch_dir = 'remote_branch'

    def setUp(self):
        super().setUp()
        if self._will_escape_isolation(self.transport_server):
            # FIXME: Some policy search ends up above the user home directory
            # and are seen as attemps to escape test isolation
            raise tests.TestNotApplicable('Escaping test isolation')
        self.remote_branch_url = self.make_remote_branch_without_working_tree()

    @staticmethod
    def _will_escape_isolation(transport_server):
        if not features.paramiko.available():
            return False
        from ....tests import stub_sftp
        if transport_server is stub_sftp.SFTPHomeDirServer:
            return True
        return False

    def make_remote_branch_without_working_tree(self):
        """Creates a branch without working tree to upload from.

        It's created from the existing self.branch_dir one which still has its
        working tree.
        """
        self.make_branch_and_working_tree()
        self.add_file('hello', b'foo')

        remote_branch_url = self.get_url(self.remote_branch_dir)
        self.run_bzr(['push', remote_branch_url,
                      '--directory', self.branch_dir])
        return remote_branch_url

    def test_no_upload_to_remote_working_tree(self):
        cmd = self._get_cmd_upload()
        up_url = self.get_url(self.branch_dir)
        # Let's try to upload from the just created remote branch into the
        # branch (which has a working tree).
        self.assertRaises(cmds.CannotUploadToWorkingTree,
                          cmd.run, up_url, directory=self.remote_branch_url)

    def test_upload_without_working_tree(self):
        self.do_full_upload(directory=self.remote_branch_url)
        self.assertUpFileEqual(b'foo', 'hello')


class TestUploadDiverged(tests.TestCaseWithTransport, UploadUtilsMixin):

    def setUp(self):
        super().setUp()
        self.diverged_tree = self.make_diverged_tree_and_upload_location()

    def make_diverged_tree_and_upload_location(self):
        tree_a = self.make_branch_and_tree('tree_a')
        tree_a.commit('message 1', rev_id=b'rev1')
        tree_a.commit('message 2', rev_id=b'rev2a')
        tree_b = tree_a.controldir.sprout('tree_b').open_workingtree()
        uncommit.uncommit(tree_b.branch, tree=tree_b)
        tree_b.commit('message 2', rev_id=b'rev2b')
        # upload tree a
        self.do_full_upload(directory=tree_a.basedir)
        return tree_b

    def assertRevidUploaded(self, revid):
        t = self.get_transport(self.upload_dir)
        uploaded_revid = t.get_bytes('.bzr-upload.revid')
        self.assertEqual(revid, uploaded_revid)

    def test_cant_upload_diverged(self):
        self.assertRaises(cmds.DivergedUploadedTree,
                          self.do_incremental_upload,
                          directory=self.diverged_tree.basedir)
        self.assertRevidUploaded(b'rev2a')

    def test_upload_diverged_with_overwrite(self):
        self.do_incremental_upload(directory=self.diverged_tree.basedir,
                                   overwrite=True)
        self.assertRevidUploaded(b'rev2b')


class TestUploadBadRemoteReivd(tests.TestCaseWithTransport, UploadUtilsMixin):

    def test_raises_on_wrong_revid(self):
        tree = self.make_branch_and_working_tree()
        self.do_full_upload()
        # Put a fake revid on the remote branch
        t = self.get_transport(self.upload_dir)
        t.put_bytes('.bzr-upload.revid', b'fake')
        # Make a change
        self.add_file('foo', b'bar\n')
        self.assertRaises(cmds.DivergedUploadedTree, self.do_full_upload)



# Copyright (C) 2005-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import base64
import re
from io import BytesIO
from urllib.request import parse_http_list, parse_keqv_list

from .. import errors, osutils, tests, transport
from ..bzr.smart import medium
from ..transport import chroot
from . import http_server


class HTTPServerWithSmarts(http_server.HttpServer):
    """HTTPServerWithSmarts extends the HttpServer with POST methods that will
    trigger a smart server to execute with a transport rooted at the rootdir of
    the HTTP server.
    """

    def __init__(self, protocol_version=None):
        http_server.HttpServer.__init__(self, SmartRequestHandler,
                                        protocol_version=protocol_version)


class SmartRequestHandler(http_server.TestingHTTPRequestHandler):
    """Extend TestingHTTPRequestHandler to support smart client POSTs.

    XXX: This duplicates a fair bit of the logic in breezy.transport.http.wsgi.
    """

    def do_POST(self):
        """Hand the request off to a smart server instance."""
        backing = transport.get_transport_from_path(
            self.server.test_case_server._home_dir)
        chroot_server = chroot.ChrootServer(backing)
        chroot_server.start_server()
        try:
            t = transport.get_transport_from_url(chroot_server.get_url())
            self.do_POST_inner(t)
        finally:
            chroot_server.stop_server()

    def do_POST_inner(self, chrooted_transport):
        self.send_response(200)
        self.send_header("Content-type", "application/octet-stream")
        if not self.path.endswith('.bzr/smart'):
            raise AssertionError(
                'POST to path not ending in .bzr/smart: {!r}'.format(self.path))
        t = chrooted_transport.clone(self.path[:-len('.bzr/smart')])
        # if this fails, we should return 400 bad request, but failure is
        # failure for now - RBC 20060919
        data_length = int(self.headers['Content-Length'])
        # TODO: We might like to support streaming responses.  1.0 allows no
        # Content-length in this case, so for integrity we should perform our
        # own chunking within the stream.
        # 1.1 allows chunked responses, and in this case we could chunk using
        # the HTTP chunking as this will allow HTTP persistence safely, even if
        # we have to stop early due to error, but we would also have to use the
        # HTTP trailer facility which may not be widely available.
        request_bytes = self.rfile.read(data_length)
        protocol_factory, unused_bytes = (
            medium._get_protocol_factory_for_bytes(request_bytes))
        out_buffer = BytesIO()
        smart_protocol_request = protocol_factory(t, out_buffer.write, '/')
        # Perhaps there should be a SmartServerHTTPMedium that takes care of
        # feeding the bytes in the http request to the smart_protocol_request,
        # but for now it's simpler to just feed the bytes directly.
        smart_protocol_request.accept_bytes(unused_bytes)
        if not (smart_protocol_request.next_read_size() == 0):
            raise errors.SmartProtocolError(
                "not finished reading, but all data sent to protocol.")
        self.send_header("Content-Length", str(len(out_buffer.getvalue())))
        self.end_headers()
        self.wfile.write(out_buffer.getvalue())


class TestCaseWithWebserver(tests.TestCaseWithTransport):
    """A support class that provides readonly urls that are http://.

    This is done by forcing the readonly server to be an http
    one. This will currently fail if the primary transport is not
    backed by regular disk files.
    """

    # These attributes can be overriden or parametrized by daughter clasess if
    # needed, but must exist so that the create_transport_readonly_server()
    # method (or any method creating an http(s) server) can propagate it.
    _protocol_version = None
    _url_protocol = 'http'

    def setUp(self):
        super().setUp()
        self.transport_readonly_server = http_server.HttpServer

    def create_transport_readonly_server(self):
        server = self.transport_readonly_server(
            protocol_version=self._protocol_version)
        server._url_protocol = self._url_protocol
        return server


class TestCaseWithTwoWebservers(TestCaseWithWebserver):
    """A support class providing readonly urls on two servers that are http://.

    We set up two webservers to allows various tests involving
    proxies or redirections from one server to the other.
    """

    def setUp(self):
        super().setUp()
        self.transport_secondary_server = http_server.HttpServer
        self.__secondary_server = None

    def create_transport_secondary_server(self):
        """Create a transport server from class defined at init.

        This is mostly a hook for daughter classes.
        """
        server = self.transport_secondary_server(
            protocol_version=self._protocol_version)
        server._url_protocol = self._url_protocol
        return server

    def get_secondary_server(self):
        """Get the server instance for the secondary transport."""
        if self.__secondary_server is None:
            self.__secondary_server = self.create_transport_secondary_server()
            self.start_server(self.__secondary_server)
        return self.__secondary_server

    def get_secondary_url(self, relpath=None):
        base = self.get_secondary_server().get_url()
        return self._adjust_url(base, relpath)

    def get_secondary_transport(self, relpath=None):
        t = transport.get_transport_from_url(self.get_secondary_url(relpath))
        self.assertTrue(t.is_readonly())
        return t


class ProxyServer(http_server.HttpServer):
    """A proxy test server for http transports."""

    proxy_requests = True


class RedirectRequestHandler(http_server.TestingHTTPRequestHandler):
    """Redirect all request to the specified server"""

    def parse_request(self):
        """Redirect a single HTTP request to another host"""
        valid = http_server.TestingHTTPRequestHandler.parse_request(self)
        if valid:
            tcs = self.server.test_case_server
            code, target = tcs.is_redirected(self.path)
            if code is not None and target is not None:
                # Redirect as instructed
                self.send_response(code)
                self.send_header('Location', target)
                # We do not send a body
                self.send_header('Content-Length', '0')
                self.end_headers()
                return False  # The job is done
            else:
                # We leave the parent class serve the request
                pass
        return valid


class HTTPServerRedirecting(http_server.HttpServer):
    """An HttpServer redirecting to another server """

    def __init__(self, request_handler=RedirectRequestHandler,
                 protocol_version=None):
        http_server.HttpServer.__init__(self, request_handler,
                                        protocol_version=protocol_version)
        # redirections is a list of tuples (source, target, code)
        # - source is a regexp for the paths requested
        # - target is a replacement for re.sub describing where
        #   the request will be redirected
        # - code is the http error code associated to the
        #   redirection (301 permanent, 302 temporarry, etc
        self.redirections = []

    def redirect_to(self, host, port):
        """Redirect all requests to a specific host:port"""
        self.redirections = [('(.*)',
                              r'http://{}:{}\1'.format(host, port),
                              301)]

    def is_redirected(self, path):
        """Is the path redirected by this server.

        :param path: the requested relative path

        :returns: a tuple (code, target) if a matching
             redirection is found, (None, None) otherwise.
        """
        code = None
        target = None
        for (rsource, rtarget, rcode) in self.redirections:
            target, match = re.subn(rsource, rtarget, path, count=1)
            if match:
                code = rcode
                break  # The first match wins
            else:
                target = None
        return code, target


class TestCaseWithRedirectedWebserver(TestCaseWithTwoWebservers):
    """A support class providing redirections from one server to another.

    We set up two webservers to allows various tests involving
    redirections.
    The 'old' server is redirected to the 'new' server.
    """

    def setUp(self):
        super().setUp()
        # The redirections will point to the new server
        self.new_server = self.get_readonly_server()
        # The requests to the old server will be redirected to the new server
        self.old_server = self.get_secondary_server()

    def create_transport_secondary_server(self):
        """Create the secondary server redirecting to the primary server"""
        new = self.get_readonly_server()
        redirecting = HTTPServerRedirecting(
            protocol_version=self._protocol_version)
        redirecting.redirect_to(new.host, new.port)
        redirecting._url_protocol = self._url_protocol
        return redirecting

    def get_old_url(self, relpath=None):
        base = self.old_server.get_url()
        return self._adjust_url(base, relpath)

    def get_old_transport(self, relpath=None):
        t = transport.get_transport_from_url(self.get_old_url(relpath))
        self.assertTrue(t.is_readonly())
        return t

    def get_new_url(self, relpath=None):
        base = self.new_server.get_url()
        return self._adjust_url(base, relpath)

    def get_new_transport(self, relpath=None):
        t = transport.get_transport_from_url(self.get_new_url(relpath))
        self.assertTrue(t.is_readonly())
        return t


class AuthRequestHandler(http_server.TestingHTTPRequestHandler):
    """Requires an authentication to process requests.

    This is intended to be used with a server that always and
    only use one authentication scheme (implemented by daughter
    classes).
    """

    # The following attributes should be defined in the server
    # - auth_header_sent: the header name sent to require auth
    # - auth_header_recv: the header received containing auth
    # - auth_error_code: the error code to indicate auth required

    def _require_authentication(self):
        # Note that we must update test_case_server *before*
        # sending the error or the client may try to read it
        # before we have sent the whole error back.
        tcs = self.server.test_case_server
        tcs.auth_required_errors += 1
        self.send_response(tcs.auth_error_code)
        self.send_header_auth_reqed()
        # We do not send a body
        self.send_header('Content-Length', '0')
        self.end_headers()
        return

    def do_GET(self):
        if self.authorized():
            return http_server.TestingHTTPRequestHandler.do_GET(self)
        else:
            return self._require_authentication()

    def do_HEAD(self):
        if self.authorized():
            return http_server.TestingHTTPRequestHandler.do_HEAD(self)
        else:
            return self._require_authentication()


class BasicAuthRequestHandler(AuthRequestHandler):
    """Implements the basic authentication of a request"""

    def authorized(self):
        tcs = self.server.test_case_server
        if tcs.auth_scheme != 'basic':
            return False

        auth_header = self.headers.get(tcs.auth_header_recv, None)
        if auth_header:
            scheme, raw_auth = auth_header.split(' ', 1)
            if scheme.lower() == tcs.auth_scheme:
                user, password = base64.b64decode(raw_auth).split(b':')
                return tcs.authorized(user.decode('ascii'),
                                      password.decode('ascii'))

        return False

    def send_header_auth_reqed(self):
        tcs = self.server.test_case_server
        self.send_header(tcs.auth_header_sent,
                         'Basic realm="%s"' % tcs.auth_realm)


# FIXME: We could send an Authentication-Info header too when
# the authentication is succesful

class DigestAuthRequestHandler(AuthRequestHandler):
    """Implements the digest authentication of a request.

    We need persistence for some attributes and that can't be
    achieved here since we get instantiated for each request. We
    rely on the DigestAuthServer to take care of them.
    """

    def authorized(self):
        tcs = self.server.test_case_server

        auth_header = self.headers.get(tcs.auth_header_recv, None)
        if auth_header is None:
            return False
        scheme, auth = auth_header.split(None, 1)
        if scheme.lower() == tcs.auth_scheme:
            auth_dict = parse_keqv_list(parse_http_list(auth))

            return tcs.digest_authorized(auth_dict, self.command)

        return False

    def send_header_auth_reqed(self):
        tcs = self.server.test_case_server
        header = 'Digest realm="%s", ' % tcs.auth_realm
        header += 'nonce="{}", algorithm="{}", qop="auth"'.format(tcs.auth_nonce,
                                                              'MD5')
        self.send_header(tcs.auth_header_sent, header)


class DigestAndBasicAuthRequestHandler(DigestAuthRequestHandler):
    """Implements a digest and basic authentication of a request.

    I.e. the server proposes both schemes and the client should choose the best
    one it can handle, which, in that case, should be digest, the only scheme
    accepted here.
    """

    def send_header_auth_reqed(self):
        tcs = self.server.test_case_server
        self.send_header(tcs.auth_header_sent,
                         'Basic realm="%s"' % tcs.auth_realm)
        header = 'Digest realm="%s", ' % tcs.auth_realm
        header += 'nonce="{}", algorithm="{}", qop="auth"'.format(tcs.auth_nonce,
                                                              'MD5')
        self.send_header(tcs.auth_header_sent, header)


class AuthServer(http_server.HttpServer):
    """Extends HttpServer with a dictionary of passwords.

    This is used as a base class for various schemes which should
    all use or redefined the associated AuthRequestHandler.

    Note that no users are defined by default, so add_user should
    be called before issuing the first request.
    """

    # The following attributes should be set dy daughter classes
    # and are used by AuthRequestHandler.
    auth_header_sent = None
    auth_header_recv = None
    auth_error_code = None
    auth_realm = "Thou should not pass"

    def __init__(self, request_handler, auth_scheme,
                 protocol_version=None):
        http_server.HttpServer.__init__(self, request_handler,
                                        protocol_version=protocol_version)
        self.auth_scheme = auth_scheme
        self.password_of = {}
        self.auth_required_errors = 0

    def add_user(self, user, password):
        """Declare a user with an associated password.

        password can be empty, use an empty string ('') in that
        case, not None.
        """
        self.password_of[user] = password

    def authorized(self, user, password):
        """Check that the given user provided the right password"""
        expected_password = self.password_of.get(user, None)
        return expected_password is not None and password == expected_password


# FIXME: There is some code duplication with
# _urllib2_wrappers.py.DigestAuthHandler. If that duplication
# grows, it may require a refactoring. Also, we don't implement
# SHA algorithm nor MD5-sess here, but that does not seem worth
# it.
class DigestAuthServer(AuthServer):
    """A digest authentication server"""

    auth_nonce = 'now!'

    def __init__(self, request_handler, auth_scheme,
                 protocol_version=None):
        AuthServer.__init__(self, request_handler, auth_scheme,
                            protocol_version=protocol_version)

    def digest_authorized(self, auth, command):
        nonce = auth['nonce']
        if nonce != self.auth_nonce:
            return False
        realm = auth['realm']
        if realm != self.auth_realm:
            return False
        user = auth['username']
        if user not in self.password_of:
            return False
        algorithm = auth['algorithm']
        if algorithm != 'MD5':
            return False
        qop = auth['qop']
        if qop != 'auth':
            return False

        password = self.password_of[user]

        # Recalculate the response_digest to compare with the one
        # sent by the client
        A1 = ('{}:{}:{}'.format(user, realm, password)).encode('utf-8')
        A2 = ('{}:{}'.format(command, auth['uri'])).encode('utf-8')

        def H(x):
            return osutils.md5(x).hexdigest()

        def KD(secret, data):
            return H(("{}:{}".format(secret, data)).encode('utf-8'))

        nonce_count = int(auth['nc'], 16)

        ncvalue = '%08x' % nonce_count

        cnonce = auth['cnonce']
        noncebit = '{}:{}:{}:{}:{}'.format(nonce, ncvalue, cnonce, qop, H(A2))
        response_digest = KD(H(A1), noncebit)

        return response_digest == auth['response']


class HTTPAuthServer(AuthServer):
    """An HTTP server requiring authentication"""

    def init_http_auth(self):
        self.auth_header_sent = 'WWW-Authenticate'
        self.auth_header_recv = 'Authorization'
        self.auth_error_code = 401


class ProxyAuthServer(AuthServer):
    """A proxy server requiring authentication"""

    def init_proxy_auth(self):
        self.proxy_requests = True
        self.auth_header_sent = 'Proxy-Authenticate'
        self.auth_header_recv = 'Proxy-Authorization'
        self.auth_error_code = 407


class HTTPBasicAuthServer(HTTPAuthServer):
    """An HTTP server requiring basic authentication"""

    def __init__(self, protocol_version=None):
        HTTPAuthServer.__init__(self, BasicAuthRequestHandler, 'basic',
                                protocol_version=protocol_version)
        self.init_http_auth()


class HTTPDigestAuthServer(DigestAuthServer, HTTPAuthServer):
    """An HTTP server requiring digest authentication"""

    def __init__(self, protocol_version=None):
        DigestAuthServer.__init__(self, DigestAuthRequestHandler, 'digest',
                                  protocol_version=protocol_version)
        self.init_http_auth()


class HTTPBasicAndDigestAuthServer(DigestAuthServer, HTTPAuthServer):
    """An HTTP server requiring basic or digest authentication"""

    def __init__(self, protocol_version=None):
        DigestAuthServer.__init__(self, DigestAndBasicAuthRequestHandler,
                                  'basicdigest',
                                  protocol_version=protocol_version)
        self.init_http_auth()
        # We really accept Digest only
        self.auth_scheme = 'digest'


class ProxyBasicAuthServer(ProxyAuthServer):
    """A proxy server requiring basic authentication"""

    def __init__(self, protocol_version=None):
        ProxyAuthServer.__init__(self, BasicAuthRequestHandler, 'basic',
                                 protocol_version=protocol_version)
        self.init_proxy_auth()


class ProxyDigestAuthServer(DigestAuthServer, ProxyAuthServer):
    """A proxy server requiring basic authentication"""

    def __init__(self, protocol_version=None):
        ProxyAuthServer.__init__(self, DigestAuthRequestHandler, 'digest',
                                 protocol_version=protocol_version)
        self.init_proxy_auth()


class ProxyBasicAndDigestAuthServer(DigestAuthServer, ProxyAuthServer):
    """An proxy server requiring basic or digest authentication"""

    def __init__(self, protocol_version=None):
        DigestAuthServer.__init__(self, DigestAndBasicAuthRequestHandler,
                                  'basicdigest',
                                  protocol_version=protocol_version)
        self.init_proxy_auth()
        # We really accept Digest only
        self.auth_scheme = 'digest'



# Copyright (C) 2008-2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#

"""Tests of the 'brz alias' command."""

from breezy import config, tests
from breezy.tests import features


class TestAlias(tests.TestCaseWithTransport):

    def test_list_alias_with_none(self):
        """Calling alias with no parameters lists existing aliases."""
        out, err = self.run_bzr('alias')
        self.assertEqual('', out)

    def test_list_unknown_alias(self):
        out, err = self.run_bzr('alias commit')
        self.assertEqual('brz alias: commit: not found\n', out)

    def test_add_alias_outputs_nothing(self):
        out, err = self.run_bzr('alias commit="commit --strict"')
        self.assertEqual('', out)

    def test_add_alias_visible(self):
        """Adding an alias makes it ..."""
        self.run_bzr('alias commit="commit --strict"')
        out, err = self.run_bzr('alias commit')
        self.assertEqual('brz alias commit="commit --strict"\n', out)

    def test_unicode_alias(self):
        """Unicode aliases should work (Bug #529930)"""
        # XXX: strictly speaking, lack of unicode filenames doesn't imply that
        # unicode command lines aren't available.
        self.requireFeature(features.UnicodeFilenameFeature)
        file_name = 'foo\xb6'

        tree = self.make_branch_and_tree('.')
        self.build_tree([file_name])
        tree.add(file_name)
        tree.commit('added')

        config.GlobalConfig.from_string(
            '[ALIASES]\nust=st {}\n'.format(file_name), save=True)

        out, err = self.run_bzr('ust')
        self.assertEqual(err, '')
        self.assertEqual(out, '')

    def test_alias_listing_alphabetical(self):
        self.run_bzr('alias commit="commit --strict"')
        self.run_bzr('alias ll="log --short"')
        self.run_bzr('alias add="add -q"')

        out, err = self.run_bzr('alias')
        self.assertEqual(
            'brz alias add="add -q"\n'
            'brz alias commit="commit --strict"\n'
            'brz alias ll="log --short"\n',
            out)

    def test_remove_unknown_alias(self):
        out, err = self.run_bzr('alias --remove fooix', retcode=3)
        self.assertEqual('brz: ERROR: The alias "fooix" does not exist.\n',
                         err)

    def test_remove_known_alias(self):
        self.run_bzr('alias commit="commit --strict"')
        out, err = self.run_bzr('alias commit')
        self.assertEqual('brz alias commit="commit --strict"\n', out)
        # No output when removing an existing alias.
        out, err = self.run_bzr('alias --remove commit')
        self.assertEqual('', out)
        # Now its not.
        out, err = self.run_bzr('alias commit')
        self.assertEqual("brz alias: commit: not found\n", out)



# Copyright (C) 2008-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for Branch.get_stacked_on_url and set_stacked_on_url."""

from breezy import branch as _mod_branch
from breezy import check, controldir, errors
from breezy.revision import NULL_REVISION
from breezy.tests import TestNotApplicable, fixtures, transport_util
from breezy.tests.per_branch import TestCaseWithBranch

unstackable_format_errors = (
    _mod_branch.UnstackableBranchFormat,
    errors.UnstackableRepositoryFormat,
    )


class TestStacking(TestCaseWithBranch):

    def check_lines_added_or_present(self, stacked_branch, revid):
        # similar to a failure seen in bug 288751 by mbp 20081120
        stacked_repo = stacked_branch.repository
        with stacked_repo.lock_read():
            list(stacked_repo.inventories.iter_lines_added_or_present_in_keys(
                [(revid,)]))

    def test_get_set_stacked_on_url(self):
        # branches must either:
        # raise UnstackableBranchFormat or
        # raise UnstackableRepositoryFormat or
        # permit stacking to be done and then return the stacked location.
        branch = self.make_branch('branch')
        target = self.make_branch('target')
        try:
            branch.set_stacked_on_url(target.base)
        except unstackable_format_errors:
            # if the set failed, so must the get
            self.assertRaises(unstackable_format_errors,
                              branch.get_stacked_on_url)
            self.assertFalse(branch._format.supports_stacking())
            return
        self.assertTrue(branch._format.supports_stacking())
        # now we have a stacked branch:
        self.assertEqual(target.base, branch.get_stacked_on_url())
        branch.set_stacked_on_url(None)
        self.assertRaises(errors.NotStacked, branch.get_stacked_on_url)

    def test_get_set_stacked_on_relative(self):
        # Branches can be stacked on other branches using relative paths.
        branch = self.make_branch('branch')
        target = self.make_branch('target')
        try:
            branch.set_stacked_on_url('../target')
        except unstackable_format_errors:
            # if the set failed, so must the get
            self.assertRaises(unstackable_format_errors,
                              branch.get_stacked_on_url)
            return
        self.assertEqual('../target', branch.get_stacked_on_url())

    def test_set_stacked_on_same_branch_raises(self):
        # Stacking on the same branch silently raises and doesn't execute the
        # change. Reported in bug 376243.
        branch = self.make_branch('branch')
        try:
            self.assertRaises(errors.UnstackableLocationError,
                              branch.set_stacked_on_url, '../branch')
        except unstackable_format_errors:
            # if the set failed, so must the get
            self.assertRaises(unstackable_format_errors,
                              branch.get_stacked_on_url)
            return
        self.assertRaises(errors.NotStacked, branch.get_stacked_on_url)

    def test_set_stacked_on_same_branch_after_being_stacked_raises(self):
        # Stacking on the same branch silently raises and doesn't execute the
        # change.
        branch = self.make_branch('branch')
        target = self.make_branch('target')
        try:
            branch.set_stacked_on_url('../target')
        except unstackable_format_errors:
            # if the set failed, so must the get
            self.assertRaises(unstackable_format_errors,
                              branch.get_stacked_on_url)
            return
        self.assertRaises(errors.UnstackableLocationError,
                          branch.set_stacked_on_url, '../branch')
        self.assertEqual('../target', branch.get_stacked_on_url())

    def assertRevisionInRepository(self, repo_path, revid):
        """Check that a revision is in a repository, disregarding stacking."""
        repo = controldir.ControlDir.open(repo_path).open_repository()
        self.assertTrue(repo.has_revision(revid))

    def assertRevisionNotInRepository(self, repo_path, revid):
        """Check that a revision is not in a repository, disregarding stacking."""
        repo = controldir.ControlDir.open(repo_path).open_repository()
        self.assertFalse(repo.has_revision(revid))

    def test_get_graph_stacked(self):
        """A stacked repository shows the graph of its parent."""
        trunk_tree = self.make_branch_and_tree('mainline')
        trunk_revid = trunk_tree.commit('mainline')
        # make a new branch, and stack on the existing one.  we don't use
        # sprout(stacked=True) here because if that is buggy and copies data
        # it would cause a false pass of this test.
        new_branch = self.make_branch('new_branch')
        try:
            new_branch.set_stacked_on_url(trunk_tree.branch.base)
        except unstackable_format_errors as e:
            raise TestNotApplicable(e)
        # reading the graph from the stacked branch's repository should see
        # data from the stacked-on branch
        new_repo = new_branch.repository
        with new_repo.lock_read():
            self.assertEqual(new_repo.get_parent_map([trunk_revid]),
                             {trunk_revid: (NULL_REVISION, )})

    def test_sprout_stacked(self):
        # We have a mainline
        trunk_tree = self.make_branch_and_tree('mainline')
        trunk_revid = trunk_tree.commit('mainline')
        # and make branch from it which is stacked
        try:
            new_dir = trunk_tree.controldir.sprout('newbranch', stacked=True)
        except unstackable_format_errors as e:
            raise TestNotApplicable(e)
        # stacked repository
        self.assertRevisionNotInRepository('newbranch', trunk_revid)
        tree = new_dir.open_branch().create_checkout('local')
        new_branch_revid = tree.commit('something local')
        self.assertRevisionNotInRepository(
            trunk_tree.branch.base, new_branch_revid)
        self.assertRevisionInRepository('newbranch', new_branch_revid)

    def test_sprout_stacked_from_smart_server(self):
        # We have a mainline
        trunk_tree = self.make_branch_and_tree('mainline')
        trunk_revid = trunk_tree.commit('mainline')
        # Make sure that we can make a stacked branch from it
        try:
            trunk_tree.controldir.sprout('testbranch', stacked=True)
        except unstackable_format_errors as e:
            raise TestNotApplicable(e)
        # Now serve the original mainline from a smart server
        remote_transport = self.make_smart_server('mainline')
        remote_bzrdir = controldir.ControlDir.open_from_transport(
            remote_transport)
        # and make branch from the smart server which is stacked
        new_dir = remote_bzrdir.sprout('newbranch', stacked=True)
        # stacked repository
        self.assertRevisionNotInRepository('newbranch', trunk_revid)
        tree = new_dir.open_branch().create_checkout('local')
        new_branch_revid = tree.commit('something local')
        self.assertRevisionNotInRepository(trunk_tree.branch.user_url,
                                           new_branch_revid)
        self.assertRevisionInRepository('newbranch', new_branch_revid)

    def test_unstack_fetches(self):
        """Removing the stacked-on branch pulls across all data"""
        try:
            builder = self.make_branch_builder('trunk')
        except errors.UninitializableFormat:
            raise TestNotApplicable('uninitializeable format')
        # We have a mainline
        trunk, mainline_revid, rev2 = fixtures.build_branch_with_non_ancestral_rev(
            builder)
        # and make branch from it which is stacked (with no tags)
        try:
            new_dir = trunk.controldir.sprout(
                self.get_url('newbranch'), stacked=True)
        except unstackable_format_errors as e:
            raise TestNotApplicable(e)
        # stacked repository
        self.assertRevisionNotInRepository('newbranch', mainline_revid)
        # TODO: we'd like to commit in the stacked repository; that requires
        # some care (maybe a BranchBuilder) if it's remote and has no
        # workingtree
        # newbranch_revid = new_dir.open_workingtree().commit('revision in '
        # 'newbranch')
        # now when we unstack that should implicitly fetch, to make sure that
        # the branch will still work
        new_branch = new_dir.open_branch()
        try:
            new_branch.tags.set_tag('tag-a', rev2)
        except errors.TagsNotSupported:
            tags_supported = False
        else:
            tags_supported = True
        new_branch.set_stacked_on_url(None)
        self.assertRevisionInRepository('newbranch', mainline_revid)
        # of course it's still in the mainline
        self.assertRevisionInRepository('trunk', mainline_revid)
        if tags_supported:
            # the tagged revision in trunk is now in newbranch too
            self.assertRevisionInRepository('newbranch', rev2)
        # and now we're no longer stacked
        self.assertRaises(errors.NotStacked, new_branch.get_stacked_on_url)

    def test_unstack_already_locked(self):
        """Removing the stacked-on branch with an already write-locked branch
        works.

        This was bug 551525.
        """
        try:
            stacked_bzrdir = self.make_stacked_bzrdir()
        except unstackable_format_errors as e:
            raise TestNotApplicable(e)
        stacked_branch = stacked_bzrdir.open_branch()
        stacked_branch.lock_write()
        stacked_branch.set_stacked_on_url(None)
        stacked_branch.unlock()

    def test_unstack_already_multiple_locked(self):
        """Unstacking a branch preserves the lock count (even though it
        replaces the br.repository object).

        This is a more extreme variation of test_unstack_already_locked.
        """
        try:
            stacked_bzrdir = self.make_stacked_bzrdir()
        except unstackable_format_errors as e:
            raise TestNotApplicable(e)
        stacked_branch = stacked_bzrdir.open_branch()
        stacked_branch.lock_write()
        stacked_branch.lock_write()
        stacked_branch.lock_write()
        stacked_branch.set_stacked_on_url(None)
        stacked_branch.unlock()
        stacked_branch.unlock()
        stacked_branch.unlock()

    def make_stacked_bzrdir(self, in_directory=None):
        """Create a stacked branch and return its bzrdir.

        :param in_directory: If not None, create a directory of this
            name and create the stacking and stacked-on bzrdirs in
            this directory.
        """
        if in_directory is not None:
            self.get_transport().mkdir(in_directory)
            prefix = in_directory + '/'
        else:
            prefix = ''
        tree = self.make_branch_and_tree(prefix + 'stacked-on')
        tree.commit('Added foo')
        stacked_bzrdir = tree.branch.controldir.sprout(
            self.get_url(prefix + 'stacked'), tree.branch.last_revision(),
            stacked=True)
        return stacked_bzrdir

    def test_clone_from_stacked_branch_preserve_stacking(self):
        # We can clone from the bzrdir of a stacked branch. If
        # preserve_stacking is True, the cloned branch is stacked on the
        # same branch as the original.
        try:
            stacked_bzrdir = self.make_stacked_bzrdir()
        except unstackable_format_errors as e:
            raise TestNotApplicable(e)
        cloned_bzrdir = stacked_bzrdir.clone('cloned', preserve_stacking=True)
        try:
            self.assertEqual(
                stacked_bzrdir.open_branch().get_stacked_on_url(),
                cloned_bzrdir.open_branch().get_stacked_on_url())
        except unstackable_format_errors as e:
            pass

    def test_clone_from_branch_stacked_on_relative_url_preserve_stacking(self):
        # If a branch's stacked-on url is relative, we can still clone
        # from it with preserve_stacking True and get a branch stacked
        # on an appropriately adjusted relative url.
        try:
            stacked_bzrdir = self.make_stacked_bzrdir(in_directory='dir')
        except unstackable_format_errors as e:
            raise TestNotApplicable(e)
        stacked_bzrdir.open_branch().set_stacked_on_url('../stacked-on')
        cloned_bzrdir = stacked_bzrdir.clone(
            self.get_url('cloned'), preserve_stacking=True)
        self.assertEqual(
            '../dir/stacked-on',
            cloned_bzrdir.open_branch().get_stacked_on_url())

    def test_clone_from_stacked_branch_no_preserve_stacking(self):
        try:
            stacked_bzrdir = self.make_stacked_bzrdir()
        except unstackable_format_errors as e:
            # not a testable combination.
            raise TestNotApplicable(e)
        cloned_unstacked_bzrdir = stacked_bzrdir.clone('cloned-unstacked',
                                                       preserve_stacking=False)
        unstacked_branch = cloned_unstacked_bzrdir.open_branch()
        self.assertRaises((errors.NotStacked, _mod_branch.UnstackableBranchFormat),
                          unstacked_branch.get_stacked_on_url)

    def test_no_op_preserve_stacking(self):
        """With no stacking, preserve_stacking should be a no-op."""
        branch = self.make_branch('source')
        cloned_bzrdir = branch.controldir.clone(
            'cloned', preserve_stacking=True)
        self.assertRaises((errors.NotStacked, _mod_branch.UnstackableBranchFormat),
                          cloned_bzrdir.open_branch().get_stacked_on_url)

    def make_stacked_on_matching(self, source):
        if source.repository.supports_rich_root():
            if source.repository._format.supports_chks:
                format = "2a"
            else:
                format = "1.9-rich-root"
        else:
            format = "1.9"
        return self.make_branch('stack-on', format)

    def test_sprout_stacking_policy_handling(self):
        """Obey policy where possible, ignore otherwise."""
        if self.bzrdir_format.fixed_components:
            raise TestNotApplicable('Branch format 4 does not autoupgrade.')
        source = self.make_branch('source')
        stack_on = self.make_stacked_on_matching(source)
        parent_bzrdir = self.make_controldir('.', format='default')
        parent_bzrdir.get_config().set_default_stack_on('stack-on')
        target = source.controldir.sprout('target').open_branch()
        # When we sprout we upgrade the branch when there is a default stack_on
        # set by a config *and* the targeted branch supports stacking.
        if stack_on._format.supports_stacking():
            self.assertEqual('../stack-on', target.get_stacked_on_url())
        else:
            self.assertRaises(
                branch.UnstackableBranchFormat, target.get_stacked_on_url)

    def test_clone_stacking_policy_handling(self):
        """Obey policy where possible, ignore otherwise."""
        if self.bzrdir_format.fixed_components:
            raise TestNotApplicable('Branch format 4 does not autoupgrade.')
        source = self.make_branch('source')
        stack_on = self.make_stacked_on_matching(source)
        parent_bzrdir = self.make_controldir('.', format='default')
        parent_bzrdir.get_config().set_default_stack_on('stack-on')
        target = source.controldir.clone('target').open_branch()
        # When we clone we upgrade the branch when there is a default stack_on
        # set by a config *and* the targeted branch supports stacking.
        if stack_on._format.supports_stacking():
            self.assertEqual('../stack-on', target.get_stacked_on_url())
        else:
            self.assertRaises(
                _mod_branch.UnstackableBranchFormat, target.get_stacked_on_url)

    def test_sprout_to_smart_server_stacking_policy_handling(self):
        """Obey policy where possible, ignore otherwise."""
        if not self.branch_format.supports_leaving_lock():
            raise TestNotApplicable('Branch format is not usable via HPSS.')
        source = self.make_branch('source')
        stack_on = self.make_stacked_on_matching(source)
        parent_bzrdir = self.make_controldir('.', format='default')
        parent_bzrdir.get_config().set_default_stack_on('stack-on')
        url = self.make_smart_server('target').base
        target = source.controldir.sprout(url).open_branch()
        # When we sprout we upgrade the branch when there is a default stack_on
        # set by a config *and* the targeted branch supports stacking.
        if stack_on._format.supports_stacking():
            self.assertEqual('../stack-on', target.get_stacked_on_url())
        else:
            self.assertRaises(
                _mod_branch.UnstackableBranchFormat, target.get_stacked_on_url)

    def prepare_stacked_on_fetch(self):
        stack_on = self.make_branch_and_tree('stack-on')
        rev1 = stack_on.commit('first commit')
        try:
            stacked_dir = stack_on.controldir.sprout('stacked', stacked=True)
        except unstackable_format_errors as e:
            raise TestNotApplicable('Format does not support stacking.')
        unstacked = self.make_repository('unstacked')
        return stacked_dir.open_workingtree(), unstacked, rev1

    def test_fetch_copies_from_stacked_on(self):
        stacked, unstacked, rev1 = self.prepare_stacked_on_fetch()
        unstacked.fetch(stacked.branch.repository, rev1)
        unstacked.get_revision(rev1)

    def test_fetch_copies_from_stacked_on_and_stacked(self):
        stacked, unstacked, rev1 = self.prepare_stacked_on_fetch()
        tree = stacked.branch.create_checkout('local')
        rev2 = tree.commit('second commit')
        unstacked.fetch(stacked.branch.repository, rev2)
        unstacked.get_revision(rev1)
        unstacked.get_revision(rev2)
        self.check_lines_added_or_present(stacked.branch, rev1)
        self.check_lines_added_or_present(stacked.branch, rev2)

    def test_autopack_when_stacked(self):
        # in bzr.dev as of 20080730, autopack was reported to fail in stacked
        # repositories because of problems with text deltas spanning physical
        # repository boundaries.  however, i didn't actually get this test to
        # fail on that code. -- mbp
        # see https://bugs.launchpad.net/bzr/+bug/252821
        stack_on = self.make_branch_and_tree('stack-on')
        if not stack_on.branch._format.supports_stacking():
            raise TestNotApplicable("%r does not support stacking"
                                    % self.branch_format)
        text_lines = [b'line %d blah blah blah\n' % i for i in range(20)]
        self.build_tree_contents([('stack-on/a', b''.join(text_lines))])
        stack_on.add('a')
        stack_on.commit('base commit')
        stacked_dir = stack_on.controldir.sprout('stacked', stacked=True)
        stacked_branch = stacked_dir.open_branch()
        local_tree = stack_on.controldir.sprout('local').open_workingtree()
        for i in range(20):
            text_lines[0] = b'changed in %d\n' % i
            self.build_tree_contents([('local/a', b''.join(text_lines))])
            local_tree.commit('commit %d' % i)
            local_tree.branch.push(stacked_branch)
        stacked_branch.repository.pack()
        check.check_dwim(stacked_branch.base, False, True, True)

    def test_pull_delta_when_stacked(self):
        if not self.branch_format.supports_stacking():
            raise TestNotApplicable("%r does not support stacking"
                                    % self.branch_format)
        stack_on = self.make_branch_and_tree('stack-on')
        text_lines = [b'line %d blah blah blah\n' % i for i in range(20)]
        self.build_tree_contents([('stack-on/a', b''.join(text_lines))])
        stack_on.add('a')
        stack_on.commit('base commit')
        # make a stacked branch from the mainline
        stacked_dir = stack_on.controldir.sprout('stacked', stacked=True)
        stacked_tree = stacked_dir.open_workingtree()
        # make a second non-stacked branch from the mainline
        other_dir = stack_on.controldir.sprout('other')
        other_tree = other_dir.open_workingtree()
        text_lines[9] = b'changed in other\n'
        self.build_tree_contents([('other/a', b''.join(text_lines))])
        stacked_revid = other_tree.commit('commit in other')
        # this should have generated a delta; try to pull that across
        # bug 252821 caused a RevisionNotPresent here...
        stacked_tree.pull(other_tree.branch)
        stacked_tree.branch.repository.pack()
        check.check_dwim(stacked_tree.branch.base, False, True, True)
        self.check_lines_added_or_present(stacked_tree.branch, stacked_revid)

    def test_fetch_revisions_with_file_changes(self):
        # Fetching revisions including file changes into a stacked branch
        # works without error.
        # Make the source tree.
        src_tree = self.make_branch_and_tree('src')
        self.build_tree_contents([('src/a', b'content')])
        src_tree.add('a')
        src_tree.commit('first commit')

        # Make the stacked-on branch.
        src_tree.controldir.sprout('stacked-on')

        # Make a branch stacked on it.
        target = self.make_branch('target')
        try:
            target.set_stacked_on_url('../stacked-on')
        except unstackable_format_errors as e:
            raise TestNotApplicable('Format does not support stacking.')

        # Change the source branch.
        self.build_tree_contents([('src/a', b'new content')])
        rev2 = src_tree.commit('second commit')

        # Fetch changes to the target.
        target.fetch(src_tree.branch)
        rtree = target.repository.revision_tree(rev2)
        rtree.lock_read()
        self.addCleanup(rtree.unlock)
        self.assertEqual(b'new content', rtree.get_file_text('a'))
        self.check_lines_added_or_present(target, rev2)

    def test_transform_fallback_location_hook(self):
        # The 'transform_fallback_location' branch hook allows us to inspect
        # and transform the URL of the fallback location for the branch.
        stack_on = self.make_branch('stack-on')
        stacked = self.make_branch('stacked')
        try:
            stacked.set_stacked_on_url('../stack-on')
        except unstackable_format_errors as e:
            raise TestNotApplicable('Format does not support stacking.')
        self.get_transport().rename('stack-on', 'new-stack-on')
        hook_calls = []

        def hook(stacked_branch, url):
            hook_calls.append(url)
            return '../new-stack-on'
        _mod_branch.Branch.hooks.install_named_hook(
            'transform_fallback_location', hook, None)
        _mod_branch.Branch.open('stacked')
        self.assertEqual(['../stack-on'], hook_calls)

    def test_stack_on_repository_branch(self):
        # Stacking should work when the repo isn't co-located with the
        # stack-on branch.
        try:
            repo = self.make_repository('repo', shared=True)
        except errors.IncompatibleFormat:
            raise TestNotApplicable()
        if not repo._format.supports_nesting_repositories:
            raise TestNotApplicable()
        # Avoid make_branch, which produces standalone branches.
        bzrdir = self.make_controldir('repo/stack-on')
        try:
            b = bzrdir.create_branch()
        except errors.UninitializableFormat:
            raise TestNotApplicable()
        transport = self.get_transport('stacked')
        b.controldir.clone_on_transport(transport, stacked_on=b.base)
        # Ensure that opening the branch doesn't raise.
        _mod_branch.Branch.open(transport.base)

    def test_revision_history_of_stacked(self):
        # See <https://launchpad.net/bugs/380314>.
        stack_on = self.make_branch_and_tree('stack-on')
        rev1 = stack_on.commit('first commit')
        try:
            stacked_dir = stack_on.controldir.sprout(
                self.get_url('stacked'), stacked=True)
        except unstackable_format_errors as e:
            raise TestNotApplicable('Format does not support stacking.')
        try:
            stacked = stacked_dir.open_workingtree()
        except errors.NoWorkingTree:
            stacked = stacked_dir.open_branch().create_checkout(
                'stacked-checkout', lightweight=True)
        tree = stacked.branch.create_checkout('local')
        rev2 = tree.commit('second commit')
        # Sanity check: stacked's repo should not contain rev1, otherwise this
        # test isn't testing what it's supposed to.
        repo = stacked.branch.repository.controldir.open_repository()
        repo.lock_read()
        self.addCleanup(repo.unlock)
        self.assertEqual({}, repo.get_parent_map([rev1]))
        # revision_history should work, even though the history is spread over
        # multiple repositories.
        self.assertEqual((2, rev2), stacked.branch.last_revision_info())


class TestStackingConnections(
        transport_util.TestCaseWithConnectionHookedTransport):

    def setUp(self):
        super().setUp()
        try:
            base_tree = self.make_branch_and_tree('base',
                                                  format=self.bzrdir_format)
        except errors.UninitializableFormat as e:
            raise TestNotApplicable(e)
        stacked = self.make_branch('stacked', format=self.bzrdir_format)
        try:
            stacked.set_stacked_on_url(base_tree.branch.base)
        except unstackable_format_errors as e:
            raise TestNotApplicable(e)
        self.rev_base = base_tree.commit('first')
        stacked.set_last_revision_info(1, self.rev_base)
        stacked_relative = self.make_branch('stacked_relative',
                                            format=self.bzrdir_format)
        stacked_relative.set_stacked_on_url(base_tree.branch.user_url)
        stacked.set_last_revision_info(1, self.rev_base)
        self.start_logging_connections()

    def test_open_stacked(self):
        b = _mod_branch.Branch.open(self.get_url('stacked'))
        rev = b.repository.get_revision(self.rev_base)
        self.assertEqual(1, len(self.connections))

    def test_open_stacked_relative(self):
        b = _mod_branch.Branch.open(self.get_url('stacked_relative'))
        rev = b.repository.get_revision(self.rev_base)
        self.assertEqual(1, len(self.connections))
